#include <string.h>
#include <stdlib.h>
#include <stdio.h>  

#include "structs.h"
#include "dcj_dist.h"


//memory recycling *************************************************

//these global variables
//handle memory recycling
//for adj structs:

struct adj_struct **adj_memory;
int adj_mem_count;
int adj_mem_total;

char* itoa(int val, int base)
{
	
	static char buf[32] = {0};
	
	int i = 30;
	
	for(; val && i ; --i, val /= base)
	
		buf[i] = "0123456789abcdef"[val % base];
	
	return &buf[i+1];
	
}

//recycling acts LIFO

struct adj_struct* new_adj(int genes)
{
	int x;

	if(adj_mem_count == adj_mem_total)
	{
		adj_memory = (struct adj_struct**)realloc(adj_memory, sizeof(struct adj_struct*) * adj_mem_count * 2);
		
		for(x = adj_mem_total; x < adj_mem_total * 2; x++)
		{
			adj_memory[x] = (struct adj_struct*)alloc_adj(5, genes);
		}
		adj_mem_total *= 2;
	}
	
	return adj_memory[adj_mem_count++];
}

void recycle_adj(void)
{
	adj_memory[--adj_mem_count]->adj_count = 0;
}


//dealing with adjacencies: **************************************

struct adj_struct* alloc_adj(int adj, int genes)
{
	int x;
	struct adj_struct* out = (struct adj_struct*)malloc(sizeof(struct adj_struct));
	
	out->adj_mem = adj;
	out->adj_count = 0;
	out->num_genes = genes;
	out->name = (char*)malloc(MAX_STR_LEN * sizeof(char));
	out->name[0] = '\0';
	
	out->adj = (int*)malloc(sizeof(int) * adj * 2);
	out->visited = (int*)malloc(sizeof(int) * genes * 2);
	out->gene = (int*)malloc(sizeof(int) * genes * 2);
		
	for(x = 0; x < adj * 2; x++)
		out->adj[x] = 2*genes + 1;
	
	
	for(x = 0; x < genes * 2; x++)
	{
		out->gene[x] = -1;
		out->visited[x] = 0;
	}
	return out;
}

void free_adj(struct adj_struct* junk)
{
	free(junk->adj);
	free(junk->gene);
	free(junk->visited);
	free(junk);	
}

//add adjacency between two gene poles, or add a telomere to an adj_struct
//if g2 is 0 then we're making a telomere
void add_adj(struct adj_struct* out, int g1, int g2)
{
	int x;

	//printf("\nadd adj: %d %d %d\n", g1, g2, out->num_genes);

	//printf("\n");
	//adj_print(out);

	//if memory use up then we need to realloc more
	if(out->adj_mem == out->adj_count)
	{
		out->adj_mem = out->adj_mem * 2;
		out->adj = (int*)realloc(out->adj, sizeof(int) * out->adj_mem * 2);
		
		for(x = out->adj_mem >> 1; x < out->adj_mem; x++)
		{
			out->adj[2*x] = 2*out->num_genes+1;
			out->adj[2*x+1] = 2*out->num_genes+1;	
		}
	}
	
	out->adj[2*out->adj_count] = GTI(g1);
	out->gene[GTI(g1)] = 2*out->adj_count;
	if(g2 != 0)
	{
		out->adj[2*out->adj_count+1] = GTI(g2);
		out->gene[GTI(g2)] = 2*out->adj_count+1;
	}
	else
	{
		out->adj[2*out->adj_count+1] = out->num_genes*2+1;	
	}
	
	out->adj_count++;
	
	//adj_print(out);
	//printf("\n\n");
}

void remove_adj(struct adj_struct* out, int index)
{
	out->adj_count--;
	if(out->adj[out->adj_count*2] != out->num_genes*2+1)
	{
		out->gene[out->adj[out->adj_count*2]] = -1;
	}
	if(out->adj[out->adj_count*2+1] != out->num_genes*2+1)
	{
		out->gene[out->adj[out->adj_count*2+1]] = -1;
	}
	out->adj[2*index] = out->adj[out->adj_count*2];
	out->adj[2*index+1] = out->adj[out->adj_count*2+1];
	
	out->adj[out->adj_count*2] = out->num_genes*2+1;
	out->adj[out->adj_count*2+1] = out->num_genes*2+1;
}

void get_adj(struct genome* a, struct adj_struct* out)
{
	int x, c_front;
	strcpy(out->name, a->name);

	//convert adjacency list
	for(x = 0; x < a->num_marker-1; x++)
	{
		if(ABS(a->genes[x]) > a->num_gene)  //if left side is a cap
		{
			//do nothing if both sides are caps, otherwise:
			if(ABS(a->genes[x+1]) <= a->num_gene) //right side is not a cap
			{	
				//record telomere on left side
				if(a->genes[x] > 0)//if linear chr
				{
					//printf("%d, 0\n", a->genes[x+1]);
					add_adj(out, a->genes[x+1], 0);
				}
				else//if circular chr
				{
					//just save the opening gene index
					c_front = x+1;
				}

			}				
		}
		else  //left side is not a cap
		{
			if(ABS(a->genes[x+1]) > a->num_gene) //right side is a cap
			{
				//record telomere on right side
				if(a->genes[x+1] > 0)
				{
					//printf("%d, 0\n", -a->genes[x]);
					add_adj(out, -a->genes[x], 0);
				}
				else
				{
					//circular
					//add an adjacency between c_front and x+1;
					
					//printf("one %d, %d\n", a->genes[c_front], -a->genes[x]);
					add_adj(out, a->genes[c_front], -a->genes[x]);
				}			
			}
			else  //neither side is a cap
			{
				//record adjacency
				//printf("two %d, %d\n", -a->genes[x], a->genes[x+1]);
				add_adj(out, -a->genes[x], a->genes[x+1]);
			}				
		}
	}
}

struct genome* adj_to_genome(struct adj_struct* in)
{
	int x, q, cap = in->num_genes+1;
	int next_write = 0;
	int chr = 0;
	
	//printf("begin adj to genome\n");
	//adj_print(in);
	//fflush(stdout);
	
	struct genome* out = (struct genome*)malloc(sizeof(struct genome));
	
	out->name = in->name;
	out->genes = (int*)malloc(sizeof(int) * 3 * in->num_genes);
	
	for(x = 0; x < in->num_genes * 2; x++)
		in->visited[x] = 0;
	
	
	//first find and order all linear chromosomes with positive caps.  Then circular with negative caps.	
	for(x = 0; x < in->adj_count; x++)
	{
		//printf("mark liner cap iteration %d %d\n", x, in->adj_count);
		//fflush(stdout);
		
		int next;
		
		if(in->visited[x])
			continue;
		
		//printf("m1\n");
		//fflush(stdout);
		
		if(in->adj[2*x+1] != 2*in->num_genes + 1)
			continue;	
			
		//printf("m2\n");
		//fflush(stdout);			
		
		if(in->adj[2*x] == 2*in->num_genes + 1)
		{
			in->visited[x] = 1;
			continue;	
		}

		//printf("m3\n");
		//fflush(stdout);

		//at this point we have an adj which is the start of a telomere
		//adj[2*x] is the gene, adj[2*x+1] is a cap
		//set the visited flag.  we'll use ADJ() to traverse to the gene, then toggle, and traverse to a new adj, TOGGLE again, and set visited.
		//if new location is num_genes + 1, then finish.  if not, repeat traverse step.
		out->genes[next_write++] = cap++;
		next = 2*x;
		
		//printf("initial next: %d\n", next);
		//fflush(stdout);
		
		out->genes[next_write++] = ITG(in->adj[next]);
		in->visited[next>>1]++;
		next = in->gene[TOGGLE(in->adj[next])];

		while(in->adj[TOGGLE(next)] != 2*in->num_genes+1)
		{
			//printf("next: %d\n", next);
			//fflush(stdout);
					
			out->genes[next_write++] = ITG(in->adj[TOGGLE(next)]);
			in->visited[next>>1]++;
			next = in->gene[TOGGLE(in->adj[TOGGLE(next)])];
		} 
		
		in->visited[next>>1]++;
		
		chr++;
		out->genes[next_write++] = cap++;
		
		//printf("close iteration\n");
		//fflush(stdout);
	}
	
	//now, every unvisited adjacency is a member of a circular chromosome
	for(x = 0; x < in->adj_count; x++)
	{
		//printf("mark circular chromosome iteration\n");
		//fflush(stdout);
		
		int next;
		
		if(in->visited[x])
			continue;
		
		out->genes[next_write++] = -cap;
		cap++;
		
		next = 2*x;
		
		while(!in->visited[next>>1])
		{
			out->genes[next_write++] = ITG(in->adj[TOGGLE(next)]);
			in->visited[next>>1]++;
			next = in->gene[TOGGLE(in->adj[TOGGLE(next)])];
		} 
		
		out->genes[next_write++] = -cap;
		cap++;
		chr++;
		
		//printf("close iteration\n");
		//fflush(stdout);
	}
		
	out->num_marker = cap-1;
	out->num_chr = chr;
   out->num_gene = in->num_genes;

	return (out);
}

void copy_adj(struct adj_struct* in, struct adj_struct* out)
{
	int x;

	//printf("copying\n");
	//adj_print(in);
	//adj_print(out);
	//fflush(stdout);

	out->num_genes = in->num_genes;

	out->name = in->name;

	out->adj_count = 0;
	
	for(x = 0; x < in->num_genes*2; x++)
		out->gene[x] = -1;
	
	for(x = 0; x < in->adj_count; x++)
	{
		//printf("trouble spot: %d %d %d\n", in->adj[2*x+1],ITG(in->adj[2*x]), ITG(in->adj[2*x+1]));
		if(in->adj[2*x+1] != in->num_genes*2+1)
		{
			//printf("three %d, %d\n", ITG(in->adj[2*x]), ITG(in->adj[2*x+1]));
			add_adj(out, ITG(in->adj[2*x]), ITG(in->adj[2*x+1]));
		}
		else
		{
			//printf("%d, %d\n", ITG(in->adj[2*x]), 0);
			add_adj(out, ITG(in->adj[2*x]), 0);
		}
	}
	
	//printf("done with that\n");
	//adj_print(in);
	//adj_print(out);
	//fflush(stdout);
}


void adj_print(struct adj_struct* a)
{
	int x;
	
	printf("name: %s mem: %d count: %d\n", a->name, a->adj_mem, a->adj_count);
	
	printf("genes: %d\ngene: \n", a->num_genes);
	for(x = 0; x < 2*a->num_genes; x++)
	{
		printf("%d ", a->gene[x]);	
	}
	printf("\nadj: \n");

	
	for(x = 0; x < a->adj_mem * 2; x++)
	{
		printf("%d, ", a->adj[x]);	
	}
	printf("\n");
}

void order_print(FILE* out, struct genome* in)
{
	int x;
	
	fprintf(out,">");
	fprintf(out,in->name);
	fprintf(out,"\n");
	
	int open = 0;
	
	for(x = 0; x < in->num_marker; x++)
	{
		if(ABS(in->genes[x]) > in->num_gene)
		{
			if(open)
			{
				open = 0;
				if(in->genes[x] > 0)
				{
					fprintf(out,"$\n");
				}
				else
				{
					fprintf(out,";\n");
				}	
			}
			else
			{
				open = 1;
			}
		}
		else
		{
			fprintf(out,"%d ", in->genes[x]);
		}
	}
	fprintf(out, "\n");
}

struct adj_struct* identity_adj(int genes)
{
	struct adj_struct* out = (struct adj_struct*)new_adj(genes);
	int x;
	
	out->name = (char*)malloc(sizeof(char)*10);
	strcpy(out->name, "Identity");
	
	add_adj(out, 1, 0);
	
	for(x = 1; x < genes; x++)
		add_adj(out, -x, x+1);
	
	add_adj(out, -genes, 0);
	
	return out;	
}

//************************************
//end adj_struct <-> MGR struct managment and conversion section
//************************************


/*int dcj_dist(struct adj_struct** in, int one, int two, struct pr** r, struct tree_controller* tc)
{
	struct adj_struct* a = in[one];
	struct adj_struct* b = in[two];
	
	int x, y, t, next, even = 0, odd = 0, cycle = 0;
	int perp_c = 0, para_c = 0, odd_c = 0;
	
	for(x = 0; x < a->num_genes*2; x++)
	{
		a->visited[x] = 0;
	}
	for(x = 0; x < b->num_genes*2; x++)
	{
		b->visited[x] = 0;
	}
	
	//find our prosthetic chromosome components
	//for each genome find missing genes and begin
	//a traversal from each,  mark visited adjacencies 
	//but don't skip them; each perp component will
	//be visited twice so divide by two
	
	//genome a
	for(x = 0; x <= a->num_genes*2; x++)
	{		
		if(a->gene[x] == -1)
		{		
			next = b->gene[x];
			
			while(1)
			{
				//parallel component
				if(next == -1)
				{
					para_c++;
					break;	
				}	

				b->visited[next>>1]++;				
				next = b->adj[TOGGLE(next)];
	
				
				//even component
				if(next == a->num_genes*2+1)
				{
					break;
				}
				
				t = next;
				next = a->gene[next];
				
				if(next == -1)
				{
					add(tc, one, ITG(x), ITG(t));
					perp_c++;
					break;	
				}

				a->visited[next>>1]++;				
				next = a->adj[TOGGLE(next)];
	
				
				if(next == a->num_genes*2+1)
				{
					odd_c++;
					break;
				}
				
				next = b->gene[next];
			}
		}
	}
	
	//genome b
	for(x = 0; x <= b->num_genes*2; x++)
	{
		if(b->gene[x] == -1)
		{	
			next = a->gene[x];
			
			while(1)
			{

				//parallel component
				if(next == -1)
				{
					break;	
				}	

				a->visited[next>>1]++;				
				next = a->adj[TOGGLE(next)];
		
				if(next == b->num_genes*2+1)
				{
					break;
				}
				
				t = next;
				next = b->gene[next];
				
				if(next == -1)
				{	
					add(tc, two, ITG(x), ITG(t));
					perp_c++;
					break;	
				}

				b->visited[next>>1]++;				
				next = b->adj[TOGGLE(next)];

				if(next == b->num_genes*2+1)
				{
					odd_c++;
					break;
				}
				
				next = a->gene[next];
			}
		}
	}	

	perp_c = perp_c>>1;

	//find every odd path telomere, and all near even path telomeres
	//by iterating through a
	for(x = 0; x < a->adj_count; x++)
	{
		//already visited
		if(a->visited[x])
			continue;
		
		//if not a telomere
		if(a->adj[2*x+1] != a->num_genes*2+1)
			continue;
		
		//if is an empty adj
		if(a->adj[2*x] == a->num_genes*2+1)
		{
			a->visited[x]++;
			continue;	
		}		
		
		//now we've found some sort of telomere; set up visited and iterator
		//traverse the first step, and start checking if subsequent steps
		//bump into another telomere

		odd++;
		next = 2*x;
		a->visited[x]++;

		next = b->gene[a->adj[next]];
		b->visited[next>>1]++;

		while(b->adj[TOGGLE(next)] != a->num_genes * 2 + 1)
		{	
			next = a->gene[b->adj[TOGGLE(next)]];
			a->visited[next>>1]++;
			
			if(a->adj[TOGGLE(next)] == a->num_genes * 2 + 1)
			{
				odd--;
				even++;
				break;
			}
	
			next = b->gene[a->adj[TOGGLE(next)]];
			b->visited[next>>1]++;
		}
	}
	
	//second, find every far even path
	for(x = 0; x < b->adj_count; x++)
	{
		//already visited
		if(b->visited[x])
			continue;
		
		//if not a telomere
		if(b->adj[2*x+1] != a->num_genes*2+1)
			continue;
		
		//if is an empty adj
		if(b->adj[2*x] == a->num_genes*2+1)
		{
			b->visited[x]++;
			continue;	
		}		
		
		//now we've found some sort of telomere; set up visited and iterator
		//traverse the first step, and start checking if subsequent steps
		//bump into another telomere

		even++;
		next = 2*x;
		b->visited[x]++;

		next = a->gene[b->adj[next]];
		a->visited[next>>1]++;

		next = b->gene[a->adj[TOGGLE(next)]];
		b->visited[next>>1]++;
		
		while(b->adj[TOGGLE(next)] != 2 * a->num_genes + 1)
		{
			next = a->gene[b->adj[TOGGLE(next)]];
			a->visited[next>>1]++;
			next = b->gene[a->adj[TOGGLE(next)]];
			b->visited[next>>1]++;		
		}	
	}
	
	//now every unvisited adj is part of a cycle
	for(x = 0; x < a->adj_count; x++)
	{
		//already visited
		if(a->visited[x])
			continue;
	
		cycle++;
		next = 2*x;
		a->visited[x]++;
		
		next = b->gene[a->adj[next]];
		b->visited[next>>1]++;

		next = a->gene[b->adj[TOGGLE(next)]];
	
		while(!a->visited[next>>1])
		{
			a->visited[next>>1]++;
			next = b->gene[a->adj[TOGGLE(next)]];
			b->visited[next>>1]++;
			next = a->gene[b->adj[TOGGLE(next)]];
		}	
	}
	
	//printf("num_g: %d even: %d odd: %d cycle: %d\n", a->num_genes, even, odd, cycle);

	(*r)->odd_path = odd;
	(*r)->cycle = cycle;
	(*r)->perp = perp_c;
	(*r)->para = para_c;
	(*r)->odd = odd_c;
	
	(*r)->lb = a->num_genes - (cycle + perp_c + para_c/2 + odd_c/2 + odd/2);
	(*r)->ub = a->num_genes - (cycle + odd/2 + para_c/2 + odd_c/2);
	
	//printf("cycle: %d odd: %d  (comps) odd: %d para: %d perp: %d\n", (*r)->cycle, (*r)->odd_path, (*r)->odd, (*r)->para, (*r)->perp);
	//printf("ub: %d lb: %d\n", (*r)->ub, (*r)->lb);
	
	return (a->num_genes - cycle) - (odd>>1);
}*/

int dcj_dist_pair_only(struct adj_struct** in, int one, int two, int* genes)
{
	struct adj_struct* a = in[one];
	struct adj_struct* b = in[two];
	
	int x, y, t, next, even = 0, odd = 0, cycle = 0;
	int perp_c = 0, para_c = 0, odd_c = 0;
	
	for(x = 0; x < a->num_genes*2; x++)
	{
		a->visited[x] = 0;
	}
	for(x = 0; x < b->num_genes*2; x++)
	{
		b->visited[x] = 0;
	}
	
	//find our prosthetic chromosome components
	//for each genome find missing genes and begin
	//a traversal from each,  mark visited adjacencies 
	//but don't skip them; each perp component will
	//be visited twice so divide by two
	
	//genome a
	for(x = 0; x <= a->num_genes*2; x++)
	{		
		if(a->gene[x] == -1)
		{		
			next = b->gene[x];
			
			while(1)
			{
				//parallel component
				if(next == -1)
				{
					para_c++;
					break;	
				}	

				b->visited[next>>1]++;				
				next = b->adj[TOGGLE(next)];
	
				
				//even component
				if(next == a->num_genes*2+1)
				{
					break;
				}
				
				t = next;
				next = a->gene[next];
				
				if(next == -1)
				{
					//add(tc, one, ITG(x), ITG(t));
					perp_c++;
					break;	
				}

				a->visited[next>>1]++;				
				next = a->adj[TOGGLE(next)];
	
				
				if(next == a->num_genes*2+1)
				{
					odd_c++;
					break;
				}
				
				next = b->gene[next];
			}
		}
	}
	
	//genome b
	for(x = 0; x <= b->num_genes*2; x++)
	{
		if(b->gene[x] == -1)
		{	
			next = a->gene[x];
			
			while(1)
			{

				//parallel component
				if(next == -1)
				{
					break;	
				}	

				a->visited[next>>1]++;				
				next = a->adj[TOGGLE(next)];
		
				if(next == b->num_genes*2+1)
				{
					break;
				}
				
				t = next;
				next = b->gene[next];
				
				if(next == -1)
				{	
					//add(tc, two, ITG(x), ITG(t));
					perp_c++;
					break;	
				}

				b->visited[next>>1]++;				
				next = b->adj[TOGGLE(next)];

				if(next == b->num_genes*2+1)
				{
					odd_c++;
					break;
				}
				
				next = a->gene[next];
			}
		}
	}	

	perp_c = perp_c>>1;

	//find every odd path telomere, and all near even path telomeres
	//by iterating through a
	for(x = 0; x < a->adj_count; x++)
	{
		//already visited
		if(a->visited[x])
			continue;
		
		//if not a telomere
		if(a->adj[2*x+1] != a->num_genes*2+1)
			continue;
		
		//if is an empty adj
		if(a->adj[2*x] == a->num_genes*2+1)
		{
			a->visited[x]++;
			continue;	
		}		
		
		//now we've found some sort of telomere; set up visited and iterator
		//traverse the first step, and start checking if subsequent steps
		//bump into another telomere

		odd++;
		next = 2*x;
		a->visited[x]++;

		next = b->gene[a->adj[next]];
		b->visited[next>>1]++;

		while(b->adj[TOGGLE(next)] != a->num_genes * 2 + 1)
		{	
			next = a->gene[b->adj[TOGGLE(next)]];
			a->visited[next>>1]++;
			
			if(a->adj[TOGGLE(next)] == a->num_genes * 2 + 1)
			{
				odd--;
				even++;
				break;
			}
	
			next = b->gene[a->adj[TOGGLE(next)]];
			b->visited[next>>1]++;
		}
	}
	
	//second, find every far even path
	for(x = 0; x < b->adj_count; x++)
	{
		//already visited
		if(b->visited[x])
			continue;
		
		//if not a telomere
		if(b->adj[2*x+1] != a->num_genes*2+1)
			continue;
		
		//if is an empty adj
		if(b->adj[2*x] == a->num_genes*2+1)
		{
			b->visited[x]++;
			continue;	
		}		
		
		//now we've found some sort of telomere; set up visited and iterator
		//traverse the first step, and start checking if subsequent steps
		//bump into another telomere

		even++;
		next = 2*x;
		b->visited[x]++;

		next = a->gene[b->adj[next]];
		a->visited[next>>1]++;

		next = b->gene[a->adj[TOGGLE(next)]];
		b->visited[next>>1]++;
		
		while(b->adj[TOGGLE(next)] != 2 * a->num_genes + 1)
		{
			next = a->gene[b->adj[TOGGLE(next)]];
			a->visited[next>>1]++;
			next = b->gene[a->adj[TOGGLE(next)]];
			b->visited[next>>1]++;		
		}	
	}
	
	//now every unvisited adj is part of a cycle
	for(x = 0; x < a->adj_count; x++)
	{
		//already visited
		if(a->visited[x])
			continue;
	
		cycle++;
		next = 2*x;
		a->visited[x]++;
		
		next = b->gene[a->adj[next]];
		b->visited[next>>1]++;

		next = a->gene[b->adj[TOGGLE(next)]];
	
		while(!a->visited[next>>1])
		{
			a->visited[next>>1]++;
			next = b->gene[a->adj[TOGGLE(next)]];
			b->visited[next>>1]++;
			next = a->gene[b->adj[TOGGLE(next)]];
		}	
	}
	
	//printf("num_g: %d even: %d odd: %d cycle: %d\n", a->num_genes, even, odd, cycle);

	/*(*r)->odd_path = odd;
	(*r)->cycle = cycle;
	(*r)->perp = perp_c;
	(*r)->para = para_c;
	(*r)->odd = odd_c;
	
	(*r)->lb = a->num_genes - (cycle + perp_c + para_c/2 + odd_c/2 + odd/2);
	(*r)->ub = a->num_genes - (cycle + odd/2 + para_c/2 + odd_c/2);*/
	
	//printf("genes: %d cycle: %d odd: %d  (comps) odd: %d para: %d perp: %d (pair) genes: %d dist: %d\n", a->num_genes, cycle, odd, odd_c, para_c, perp_c, a->num_genes - (para_c>>1), *genes + -cycle + -odd/2 + odd_c/-2 - perp_c);
	//printf("ub: %d lb: %d\n", (*r)->ub, (*r)->lb);
	
	(*genes) = a->num_genes - (para_c>>1);
	
	return *genes + -cycle + -odd/2 + odd_c/-2 - perp_c;
}

/*void dataset_bound(struct adj_struct** gl, int genomes, struct pr* report, struct tree_controller* tc)
{
	int x, y;
	
	report->odd_path = 0;
	report->cycle = 0;
	
	report->perp = 0;
	report->para = 0;
	report->odd = 0;
	
	report->lb = 0;
	report->ub = 0;
	
	for(x = 0; x < genomes-1; x++)
	{
		for(y = x+1; y < genomes; y++)
		{
			struct pr t_m;
			struct pr* t = &t_m;

			dcj_dist(gl, x, y, &t, tc);

			report->odd_path = report->odd_path + t_m.odd_path;
			report->cycle = report->cycle + t_m.cycle;
			
			report->perp = report->perp + t_m.perp;
			report->para = report->para + t_m.para;
			report->odd = report->odd + t_m.odd;
			
			report->lb = report->lb + t_m.lb;
			report->ub = report->ub + t_m.ub;	
		}
	}	
}*/

/*void search_prosthetic_v3(struct adj_struct **gl, int genomes)
{
	int LB = 0;
	
	struct adj_struct* cp = alloc_adj(gl[0]->num_genes*2, gl[0]->num_genes);
	struct tree_memory* tm = init_tree_mem(5,5);
	
	while(!sp_v3(gl, genomes, cp, LB, tm))
	{
		//printf("increasing lower bound: %d\n", LB);
		fflush(stdout);
		LB++;	
	}	
}*/

/*int sp_v3(struct adj_struct** gl, int genomes, struct adj_struct* cp, int lower_bound, struct tree_memory* tm)
{
	struct tree_controller* tc = get_tc(tm);
	struct pr report;
	int current, current_adj;
	int y, z;
	
	dataset_bound(gl, genomes, &report, tc);

	if(report.lb > lower_bound)
	{
		recycle_tc(tm);
		return 0;
	}
	//if there are no more perpendicular components then we are done
	if(tc->next == 0)
	{
		for(y = 0; y < genomes; y++)
		{
			copy_adj(gl[y], cp);
			
			for(z = 0; z < gl[y]->num_genes*2; z++)
			{
				if(cp->gene[z] == -1)
				{
					add_adj(cp, ITG(z), 0);	
				}	
			}
				
			struct genome* g = adj_to_genome(cp);
				
			order_print(stdout, g);
			fflush(stdout);
			
			free(g->genes);
			free(g);
		}
		
		recycle_tc(tm);
		return 1;
	}
	
	tally(tc);
	
	while(tc->rn < tc->next)
	{
		current = tc->ranks[tc->rn++];
		
		if(!check_prev(tm, tc->elements[6*current+2], tc->elements[6*current+3], tc->elements[6*current+4]))
			continue;
		
		current_adj = gl[tc->elements[6*current+2]]->adj_count;
		add_adj(gl[tc->elements[6*current+2]], tc->elements[6*current+3], tc->elements[6*current+4]);
			
		if(sp_v3(gl, genomes, cp, lower_bound, tm))
		{
			recycle_tc(tm);
			return 1;
		}
			
		remove_adj(gl[tc->elements[6*current+2]], current_adj);
	}
		
	recycle_tc(tm);
	
	return 0;
}*/

/*void sp_v4(struct adj_struct** gl, int genomes, struct adj_struct* cp, struct tree_memory* tm)
{
	static int best = 210000000;
	struct tree_controller* tc = get_tc(tm);
	struct pr report;
	int current, current_adj;
	int y, z;
	
	dataset_bound(gl, genomes, &report, tc);

	if(report.ub < best)
	{
		best = report.ub;
		printf("new best upper bound: %d\n", report.ub);
		fflush(stdout);
					
		for(y = 0; y < genomes; y++)
		{
			//printf("y = %d\n", y);
			//fflush(stdout);
			
			copy_adj(gl[y], cp);
			
			//printf("mark1\n");
			//fflush(stdout);
			
			for(z = 0; z < gl[y]->num_genes*2; z++)
			{
				if(cp->gene[z] == -1)
				{
					add_adj(cp, ITG(z), 0);	
				}	
			}
					
			//adj_print(cp);
			//fflush(stdout);
					
			struct genome* g = adj_to_genome(cp);
					
			//order_print(stdout, g);
			//fflush(stdout);
			
			free(g->genes);
			free(g);
		}
	}
	
	//check to possibly prune
	if(report.lb >= best)
	{
		recycle_tc(tm);
		return;
	}
	
	//if there are no more perpendicular components then we are done
	if(tc->next == 0)
	{
		recycle_tc(tm);
		return;
	}

	tally(tc);

	while(tc->rn < tc->next)
	{
	
		current = tc->ranks[tc->rn++];
		
		if(!check_prev(tm, tc->elements[6*current+2], tc->elements[6*current+3], tc->elements[6*current+4]))
			continue;
		
		current_adj = gl[tc->elements[6*current+2]]->adj_count;
		add_adj(gl[tc->elements[6*current+2]], tc->elements[6*current+3], tc->elements[6*current+4]);
			
		sp_v4(gl, genomes, cp, tm);
			
		remove_adj(gl[tc->elements[6*current+2]], current_adj);
	}
		
	recycle_tc(tm);
}*/

/*void dataset_bound(struct adj_struct** gl, int genomes, int ignore, struct pr* report)
{
	int x, y;
	
	report->odd_path = 0;
	report->cycle = 0;
	
	report->perp = 0;
	report->para = 0;
	report->odd = 0;
	
	report->lb = 0;
	report->ub = 0;
	
	for(x = 0; x < genomes-1; x++)
	{
		if(ignore == x)
			continue;
		
		for(y = x+1; y < genomes; y++)
		{
			if(ignore == y)
				continue;
			
			struct pr t_m;
			struct pr* t = &t_m;
		
			dcj_dist(gl[x], gl[y], &t);
			
			report->odd_path = report->odd_path + t_m.odd_path;
			report->cycle = report->cycle + t_m.cycle;
			
			report->perp = report->perp + t_m.perp;
			report->para = report->para + t_m.para;
			report->odd = report->odd + t_m.odd;
			
			report->lb = report->lb + t_m.lb;
			report->ub = report->ub + t_m.ub;	
		}
	}	
}*/

/*void genome_bound(struct adj_struct** gl, int genomes, int current, struct pr* report)
{
	int x, y;
	
	report->odd_path = 0;
	report->cycle = 0;
	
	report->perp = 0;
	report->para = 0;
	report->odd = 0;
	
	report->lb = 0;
	report->ub = 0;
	
	for(x = 0; x < genomes; x++)
	{
		if(x == current)
			continue;
		
		struct pr t_m;
		struct pr* t = &t_m;
		
		dcj_dist(gl[x], gl[current], &t);
			
		report->odd_path = report->odd_path + t_m.odd_path;
		report->cycle = report->cycle + t_m.cycle;
			
		report->perp = report->perp + t_m.perp;
		report->para = report->para + t_m.para;
		report->odd = report->odd + t_m.odd;
			
		report->lb = report->lb + t_m.lb;
		report->ub = report->ub + t_m.ub;	
	}	
}*/

/*void search_prosthetic(struct adj_struct** gl, int genomes, struct adj_struct* cp)
{
	int anchor = 0; //the gene terminal this layer will search on
	int genome = 0; //
	int x,y,z;
	
	static int best = 100000000;

	struct pr report;
	struct pr ds_report;
	
	int current_adj = 0;
	
	//find and set the anchor
	while(1)
	{
			if(gl[genome]->gene[anchor] == -1)
				break;
			
			anchor++;
			
			if(anchor >= 2*gl[genome]->num_genes)
			{
				anchor = 0;
				genome++;
								
				if(genome >= genomes)
				{
					return;	
				}
			}				
	}	
	
	dataset_bound(gl, genomes, genome, &ds_report);
	
	//match the anchor with each following ghost label
	for(x = anchor+1; x<2*gl[genome]->num_genes; x++)
	{
		if(gl[genome]->gene[x] == -1)
		{
		
			//printf("link with %d\n", x);
		
			current_adj = gl[genome]->adj_count;
			add_adj(gl[genome], ITG(anchor), ITG(x));
			
			genome_bound(gl, genomes, genome, &report);
			
			if(report.ub + ds_report.ub < best)
			{
				best = report.ub + ds_report.ub;
				printf("new best upper bound: %d\n", report.ub + ds_report.ub);
				fflush(stdout);
						
				for(y = 0; y < genomes; y++)
				{
					copy_adj(gl[y], cp);
					
					for(z = 0; z < gl[y]->num_genes*2; z++)
					{
						if(cp->gene[z] == -1)
						{
							add_adj(cp, ITG(z), 0);	
						}	
					}
					
					struct genome* g = adj_to_genome(cp);
					
					order_print(stdout, g);
				}
			}
			
			//if report is good, search it, otherwise don't
			if(report.lb + ds_report.lb < best)
			{
				search_prosthetic(gl, genomes, cp);
			}
			else
			{
				//printf("prune\n");
			}
			
			remove_adj(gl[genome], current_adj);
		}	
	}
	
	//cap the anchor
	current_adj = gl[genome]->adj_count;
	add_adj(gl[genome], ITG(anchor), 0);
	
	genome_bound(gl, genomes, genome, &report);
	
	//printf("cap anchor\n");
	
	//printf("ub: %d lb: %d\n", report.ub, report.lb);
	
	if(report.ub + ds_report.ub < best)
	{
		best = report.ub + ds_report.ub;
		printf("new best upper bound: %d\n", report.ub + ds_report.ub);
		fflush(stdout);
		
		copy_adj(gl[y], cp);
					
		for(z = 0; z < gl[y]->num_genes*2; z++)
		{
			if(cp->gene[z] == -1)
			{
				add_adj(cp, ITG(z), 0);	
			}	
		}
					
		struct genome* g = adj_to_genome(cp);
					
		order_print(stdout, g);
	}
	
	//if report is good, search it, otherwise don't
	if(report.lb + ds_report.lb < best)
	{
		search_prosthetic(gl, genomes, cp);
	}
	else
	{
		//printf("prune\n");	
	}
	
	remove_adj(gl[genome], current_adj);
	
}

void search_prosthetic_main(struct adj_struct **gl, int genomes)
{
	int LB = 0;
	
	struct adj_struct* cp = alloc_adj(gl[0]->num_genes*2, gl[0]->num_genes);
	
	while(!search_prosthetic_v2(gl, genomes, cp, LB))
	{
		printf("Increasing lower bound: %d\n", LB);
		fflush(stdout);
		LB++;
	}		
}

int search_prosthetic_v2(struct adj_struct** gl, int genomes, struct adj_struct* cp, int lower_bound)
{
	int x, y;
	
	int anchor = 0;
	int genome = 0;
	
	int current_adj = 0;
	
	struct pr g_report;
	struct pr ds_report;
	
	while(1)
	{
		if(gl[genome]->gene[anchor] == -1)
			break;
		
		anchor++;
		
		if(anchor >= 2*gl[genome]->num_genes)
		{
			anchor = 0;
			genome++;
							
			if(genome >= genomes)
			{
				for(y = 0; y < genomes; y++)
				{
					copy_adj(gl[y], cp);										
					struct genome* g = adj_to_genome(cp);
					order_print(stdout, g);
				}
				return 1;
			}
		}				
	}	
		
	dataset_bound(gl, genomes, genome, &ds_report);
		
	//match the anchor with each following ghost label
	for(x = anchor+1; x<2*gl[genome]->num_genes; x++)
	{
		if(gl[genome]->gene[x] == -1)
		{
			current_adj = gl[genome]->adj_count;
			add_adj(gl[genome], ITG(anchor), ITG(x));
			
			genome_bound(gl, genomes, genome, &g_report);
						
			//if report is good, search it, otherwise don't
			if(g_report.lb + ds_report.lb <= lower_bound)
			{
				if(search_prosthetic_v2(gl, genomes, cp, lower_bound))
					return 1;
			}
			
			remove_adj(gl[genome], current_adj);
		}	
	}
	
	//cap the anchor
	current_adj = gl[genome]->adj_count;
	add_adj(gl[genome], ITG(anchor), 0);
	
	genome_bound(gl, genomes, genome, &g_report);
	
	//if report is good, search it, otherwise don't
	if(g_report.lb + ds_report.lb <= lower_bound)
	{
		if(search_prosthetic_v2(gl, genomes, cp, lower_bound))
			return 1;
	}
	
	remove_adj(gl[genome], current_adj);
	
	return 0;
}*/

//*********************************************
//Begin DCJ event managment section
//*********************************************


void apply_dcj(struct adj_struct* a, struct dcj* d)
{	
	int x;
	
	//printf("d->f: %d d->s: %d other adj: %d %d\n", d->f, d->s, a->gene[GTI(d->f)], a->gene[GTI(d->s)]);
	
	if(d->f == 0) //split event
	{
		//printf("split event f\n");
		//fflush(stdout);
		
		if(a->adj[TOGGLE(a->gene[GTI(d->s)])] != a->num_genes*2+1)
		{
			int f = d->s;
			int s = ITG(a->adj[TOGGLE(a->gene[GTI(d->s)])]);
					
			remove_adj(a, GTI(f)>>1);
			
			add_adj(a, f, 0);
			add_adj(a, s, 0);
			
			//adj_print(a);
		}
		return;
	}
	
	if(d->s == 0) //split event
	{
		//printf("split event s %d\n", d->f);
		//adj_print(a);
		//printf("%d %d \n", d->f, GTI(d->f));
		//fflush(stdout);
		
		if(a->adj[TOGGLE(a->gene[GTI(d->f)])] != a->num_genes*2+1)
		{
			int f = d->f;
			int s = ITG(a->adj[TOGGLE(a->gene[GTI(d->f)])]);
			//printf("\n\n\n");
			//printf("f: %d s: %d  \n", f, s);
			//adj_print(a);
			
			remove_adj(a, a->gene[GTI(f)]>>1);
			
		//	adj_print(a);
			
			//printf("%d %d\n", f, s);
			
			add_adj(a, s, 0);
			
		//	adj_print(a);
			
			add_adj(a, f, 0);
			
		//	adj_print(a);
		}
		return;
	}
	//merge event
	if(a->adj[TOGGLE(a->gene[GTI(d->f)])] == a->num_genes*2+1 && a->adj[TOGGLE(a->gene[GTI(d->s)])] == a->num_genes*2+1)
	{
		//adj_print(a);
		//printf("merge event %d %d\n", d->f, d->s);
		//printf("%d %d\n", GTI(d->f), GTI(d->s));
		//printf("%d %d\n", TOGGLE(a->gene[GTI(d->f)]), TOGGLE(a->gene[GTI(d->s)]));
		//printf("%d %d %d %d\n",a->adj[a->gene[GTI(d->f)]],a->adj[a->gene[GTI(d->s)]],a->adj[TOGGLE(a->gene[GTI(d->f)])],a->adj[TOGGLE(a->gene[GTI(d->s)])]);
		//fflush(stdout);
		
		int f = ITG(a->adj[a->gene[GTI(d->f)]]);
		int s = ITG(a->adj[a->gene[GTI(d->s)]]);
		
		remove_adj(a, a->gene[GTI(d->f)]>>1);
		remove_adj(a, a->gene[GTI(d->s)]>>1);
		
		//printf("%d %d\n", f, s);
		add_adj(a, f, s);
		return;
	}
	
	
	//printf("swap event\n");
	//fflush(stdout);
	//swap event
	int f = a->adj[a->gene[GTI(d->f)]];
	int s = a->adj[a->gene[GTI(d->s)]];
	
	a->adj[a->gene[GTI(d->f)]] = s;
	a->adj[a->gene[GTI(d->s)]] = f;
	
	f = a->gene[GTI(d->f)];
	s = a->gene[GTI(d->s)];
	
	a->gene[GTI(d->s)] = f;
	a->gene[GTI(d->f)] = s;
}

int brute_score(struct adj_struct** adj_list, int num_genomes)
{
	int b;
	
	static int best = 100000000;
	
	int x, y, z;
	
	int save_index;
		
	int flag = 1;

	for(x = 0; x < num_genomes && flag; x++)
	{
		for(y = 0; y < adj_list[x]->num_genes*2 && flag; y++)
		{	
			if(adj_list[x]->gene[y] != -1)
				continue;
			
			for(z = y+1; z < adj_list[x]->num_genes*2; z++)
			{
			if(adj_list[x]->gene[z] != -1)
					continue;
						
				flag = 0;	
				save_index = adj_list[x]->adj_count;
				
				add_adj(adj_list[x], ITG(y), ITG(z));
				
				b = brute_score(adj_list, num_genomes);
				if(b < best)
					best = b;

				remove_adj(adj_list[x], save_index);
			}

			flag = 0;
			save_index = adj_list[x]->adj_count;
			
			add_adj(adj_list[x], ITG(y), 0);
			
			b = brute_score(adj_list, num_genomes);
			if(b < best)
				best = b;
			
			remove_adj(adj_list[x], save_index);
		}
	}

	//score data set 
	if(flag)
	{	
		b = matrix_score(adj_list, num_genomes);
		
		if(b < best)
		{
			best = b;
			//printf("new best: %d\n", best);
			
			//for(x = 0; x < num_genomes; x++)
			//	adj_print(adj_list[x]);
		}		
	}

	return best;
}

int brute_score_out(struct adj_struct** adj_list, int num_genomes, int find_score)
{
	int b, best = 100000000;
	
	int x, y, z;
	
	int save_index;
		
	int flag = 1;

	for(x = 0; x < num_genomes && flag; x++)
	{
		for(y = 0; y < adj_list[x]->num_genes*2 && flag; y++)
		{	
			if(adj_list[x]->gene[y] != -1)
				continue;
			
			for(z = y+1; z < adj_list[x]->num_genes*2; z++)
			{
				if(adj_list[x]->gene[z] != -1)
					continue;
						
				flag = 0;	
				save_index = adj_list[x]->adj_count;
				
				b = brute_score_out(adj_list, num_genomes, find_score);
				if(b < best)
					best = b;

				
			}

			flag = 0;
			save_index = adj_list[x]->adj_count;
			
			
			b = brute_score_out(adj_list, num_genomes, find_score);
			if(b < best)
				best = b;
			
			
		}
	}

	//score data set 
	if(flag)
	{	
		b = matrix_score(adj_list, num_genomes);

		if(b == find_score)
		{
			for(x = 0; x < num_genomes; x++)
			{
				adj_print(adj_list[x]);	
			}	
			printf("\n");
		}
		
		if(b < best)
			best = b;		
	}
	//return best
	return best;
}

int simple_dcj_dist(struct adj_struct* a, struct adj_struct* b)
{
	int x, next, even = 0, odd = 0, cycle = 0;
	
	for(x = 0; x < a->adj_count; x++)
	{
		a->visited[x] = 0;
	}
	for(x = 0; x < b->adj_count; x++)
	{
		b->visited[x] = 0;
	}

	//printf("\n\nbeginning brand new dcj dist\n");

	//adj_print(a);
	//adj_print(b);
	
	//first find every odd path telomere, and all near even path telomeres
	//by iterating through a
	for(x = 0; x < a->adj_count; x++)
	{	
		//already visited
		if(a->visited[x])
			continue;
		
		//if not a telomere
		if(a->adj[2*x+1] != a->num_genes*2+1)
			continue;
		
		//if is an empty adj
		if(a->adj[2*x] == a->num_genes*2+1)
		{
			//printf("visiting adj %d in A\n", x);
			a->visited[x]++;
			continue;	
		}		
		
		//now we've found some sort of telomere; set up visited and iterator
		//traverse the first step, and start checking if subsequent steps
		//bump into another telomere

		odd++;
		next = 2*x;
	//	printf("visiting adj %d in A\n", x);
		a->visited[x]++;

		next = b->gene[a->adj[next]];
	//	printf("visiting adj %d in B\n", next>>1);
		b->visited[next>>1]++;

		while(b->adj[TOGGLE(next)] != a->num_genes * 2 + 1)
		{	
			next = a->gene[b->adj[TOGGLE(next)]];
		//	printf("visiting adj %d in A\n", next>>1);
			a->visited[next>>1]++;
			
			if(a->adj[TOGGLE(next)] == a->num_genes * 2 + 1)
			{
				odd--;
				even++;
				break;
			}
	
			next = b->gene[a->adj[TOGGLE(next)]];
		//	printf("visiting adj %d in B\n", next>>1);
			b->visited[next>>1]++;
		}
	}
	
	//printf("second pass: %d\n", b->adj_count);
	//second, find every far even path
	for(x = 0; x < b->adj_count; x++)
	{	
		//already visited
		if(b->visited[x])
			continue;
		
		//if not a telomere
		if(b->adj[2*x+1] != a->num_genes*2+1)
			continue;
		
		//if is an empty adj
		if(b->adj[2*x] == a->num_genes*2+1)
		{
			//printf("visiting adj %d in B\n", x);
		
			b->visited[x]++;
			continue;	
		}		
		
		//now we've found some sort of telomere; set up visited and iterator
		//traverse the first step, and start checking if subsequent steps
		//bump into another telomere

		even++;
		next = 2*x;
		//printf("visited adj %d in B\n", x);
		b->visited[x]++;

		next = a->gene[b->adj[next]];
		//printf("visiting adj %d in A\n", next>>1);
		
		a->visited[next>>1]++;

		next = b->gene[a->adj[TOGGLE(next)]];
		//printf("visiting adj %d in B\n", next>>1);
		
		b->visited[next>>1]++;
		
		while(b->adj[TOGGLE(next)] != 2 * a->num_genes + 1)
		{
			next = a->gene[b->adj[TOGGLE(next)]];
			//printf("visiting adj %d in A\n", next>>1);
		
			a->visited[next>>1]++;
			next = b->gene[a->adj[TOGGLE(next)]];
			//printf("visiting adj %d in B\n", next>>1);
		
			b->visited[next>>1]++;		
		}	
	}
	
	//printf("final pass:\n");
	//now every unvisited adj is part of a cycle
	for(x = 0; x < a->adj_count; x++)
	{	
		//already visited
		if(a->visited[x])
			continue;
	
		cycle++;
		next = 2*x;
		//printf("visiting adj %d in A\n", x);
		
		a->visited[x]++;
		
		//printf("mark next a->adj[%d]\n", next);
		
		next = b->gene[a->adj[next]];
		//printf("visiting adj %d in B\n", next>>1);
		
		b->visited[next>>1]++;

		//printf("mark0 next b->adj[toggle(%d)]\n", next);

		next = a->gene[b->adj[TOGGLE(next)]];
	
		//printf("mark1 next a->adj[toggle(%d)]\n", next);
	
		while(!a->visited[next>>1])
		{		
			//fflush(stdout);
			//printf("visiting adj %d in A\n", next>>1);
			a->visited[next>>1]++;
			next = b->gene[a->adj[TOGGLE(next)]];
			
			//if(next == 51)
				//printf("signpost!\n");
			
			//printf("mark 2 next b->adj[toggle(%d)]\n", next);
			
			//if(next == 51)
			//{
				//printf("mark4.5 %d\n", next);
				//adj_print(a);
				//adj_print(b);
			//}
		//	printf("visiting adj %d in B\n", next>>1);
		
			b->visited[next>>1]++;
			
			next = a->gene[b->adj[TOGGLE(next)]];
			//printf("mark 3 next a->adj[toggle(%d)]\n", next);
		}	
	}
	

	//printf("num_g: %d even: %d odd: %d cycle: %d\n", a->num_genes, even, odd, cycle);

	
	return (a->num_genes - cycle) - (odd>>1);
}

int matrix_score(struct adj_struct** adj_list, int num_genomes)
{
	int x, y;
	int sum = 0;
	
	for(x = 0; x < num_genomes-1; x++)
	{
		for(y = x+1; y < num_genomes; y++)
		{
			//printf("before %d %d\n", x, y);
			//adj_print(adj_list[3]);
			sum+=simple_dcj_dist(adj_list[x], adj_list[y]);
			//printf("after\n");
			//adj_print(adj_list[3]);
		}	
	}	
	
	return sum;
}
