#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "structs.h"
#include "input.h"

//read input for unequal content dataset without duplicate genes

//uses a state machine with 15 states

void read_data(FILE* fp, struct genome** output, int* num_genomes)
{
	int x, y;
	
	//lookup table information
	int genome_count = 0;
	
	int gene_max = 0;
	int gene_mem = 8;
	
	int* table;
	
	//second pass variables
	int chr_count;
	int current_genome;
	int write_location;
	int current_cap;
	int chr_front_cap;
	
	//state machine variables
	int state = 1;
	int line = 1;
	char c;

	char* digit = (char*)malloc(sizeof(char) * MAX_STR_LEN);
	
	int gene;
	
	int name_c = 0;

	
	int digit_c = 0;
	digit[0] = '\0';
	
	//setup table inital conditions
	table = (int*)malloc(sizeof(int) * gene_mem);
	
	for(x = 0; x < gene_mem; x++)
	{
		table[x] = -1;	
	}
	
	
	//first pass, establish memory sizes, build lookups to check for duplicates/missing genes
	while (1) 
  	{
		c=getc(fp);
		
		if(c == EOF)
		{
			break;
		}
    	
    	//waiting for first genome marker
    	if(state == 1)
    	{
    		if(c == '>')
    		{	
    			genome_count++;
    			state = 16;	
    			continue;
    		}
    		if(c == '#')
    		{
    			state = 12;
    			continue;	
    		}
    		if(c == '\n' || isspace(c))
    		{
    			continue;	
    		}

  			fprintf(stderr,"ERROR: \tExpecting '>' to mark initial genome or a comment.\n");
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//comment at beginning of file
    	if(state == 12)
    	{
    		if(c == '\n')
    		{
    			state = 1;
    			line++;
    		}
    		
    		continue;	
    	}
    	
    	//make sure name has at least one character
    	if(state == 16)
    	{
    		if(!isspace(c) && c != '\n')
    		{
    			state = 2;
    			continue;
    		}
    		
    		fprintf(stderr,"ERROR line %d: \tGenome name must have at least one character.\n", line);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//reading genome name characters
    	if(state == 2)
    	{
    		if(c == '\n')
    		{
    			line++;
    			state = 4;
    			continue;
    		}
    		
    		if(isspace(c))
    		{
    			state = 3;
    			continue;
    		}
    		
    		if(c == '#')
    		{
    			state = 13;
    			continue;	
    		}
    		
    		if(name_c == MAX_STR_LEN)
    		{
    			fprintf(stderr,"ERROR line %d: \tMaximum genome name size exceeded.\n", line);
     			fprintf(stderr,"\tInput file requires modification.");
     			exit(-1);
    		}
    		
    		continue;	
    	}
    	
    	//whitespace after genome name
    	if(state == 3)
    	{
    		if(c == '#')
    		{
    			state = 13;
    			continue;	
    		}
    		
    		if(c == '\n')
    		{
    			line++;
    			state = 4;
    			continue;	
    		}
    		
    		if(isspace(c))
    		{
    			continue;	
    		}
    		
    		fprintf(stderr,"ERROR line %d: \tWhitespace not permitted in genome name.\n", line);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//comment after genome name
    	if(state == 13)
    	{
    		if(c == '\n')
    		{
    			line++;
    			state = 4;
    			continue;	
    		}	
    		
    		continue;
    	}
    	
    	//begin reading first gene in first chromosome of genome
    	if(state == 4)
    	{
    	   if(c == '-')
    	   {
    	   	state = 6;
    	   	continue;
    	   }
    	   if(isdigit(c))
    	   {
    	   	state = 5;
    	   	
    	   	digit_c++;
    	   	digit[0] = c;
    	   	digit[1] = '\0';
    	   	
    	   	continue;
    	   }	
    	   if(c == '#')
    	   {
    	   	state = 14;
    	   	continue;	
    	   }
    	   	
    	   if(isspace(c))
    	   	continue;
    	   	
    		fprintf(stderr,"ERROR line %d: \tExpecting comment or '-' or digit to mark initial gene.\n", line);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//ignoring comment on a line after a genome name
    	if(state == 14)
    	{
    		if(c == '\n')
    		{
    			line++;
    			state = 4;
    			continue;	
    		}	
    		
    		continue;
    	}
    	
    	//read positive orientation gene
    	if(state == 5)
    	{
    		if(isdigit(c))
    		{
    			digit[digit_c++] = c;
    			digit[digit_c] = '\0';
    			
    			continue;
    		}
    		
    		if(c == '$')
    		{
    			//write gene
    			gene = atoi(digit);
    			    			
    			if(gene > gene_max)
    				gene_max = gene;
    			
    			//need to expand gene_mem table
    			while(gene >= gene_mem)
    			{
    				table = (int*)realloc(table, sizeof(int) * gene_mem * 2);
    				for(x = gene_mem; x < gene_mem * 2; x++)
    					table[x] = -1;
    				
    				gene_mem*=2;
    			}
    			
    			if(table[gene] == -1)
    				table[gene] = 1;
    			else
    				table[gene]++;
    			
    			
    			digit[0] = '\0';
    			digit_c = 0;
    			
    			//linear chromosome
    			state = 9;
    			continue;
    		}
    		
    		if(c == ';')
    		{
    			//write gene
    			gene = atoi(digit);

     			
    			if(gene > gene_max)
    				gene_max = gene;
    		   			
    			//need to expand gene_mem table
    			while(gene >= gene_mem)
    			{
    				table = (int*)realloc(table, sizeof(int) * gene_mem * 2);
    				for(x = gene_mem; x < gene_mem * 2; x++)
    					table[x] = -1;
    				
    				gene_mem*=2;
    			}
    			
    			if(table[gene] == -1)
    				table[gene] = 1;
    			else
    				table[gene]++;
				
    			    		
    			digit[0] = '\0';
    			digit_c = 0;
    			
    			//circular chromosome	
    			state = 9;
    			continue;
    		}
    		
    		if(isspace(c))
    		{
   			
    			//write gene
    			gene = atoi(digit);

    			
    			if(gene > gene_max)
    				gene_max = gene;
    		   			
    			//need to expand gene_mem table
    			while(gene >= gene_mem)
    			{
    				table = (int*)realloc(table, sizeof(int) * gene_mem * 2);
    				for(x = gene_mem; x < gene_mem * 2; x++)
    					table[x] = -1;
    				
    				gene_mem*=2;
    			}
    			
   			
    			if(table[gene] == -1)
    				table[gene] = 1;
    			else
    				table[gene]++;
    			
    				        			
    			digit[0] = '\0';
    			digit_c = 0;
    			
   			
    			state = 8;
    			continue;
    		}
    		
    		fprintf(stderr,"ERROR line %d: \t Invalid character %c inside chromosome.\n", line, c);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//read first digit of negative orientation gene
    	if(state == 6)
    	{
    		if(isdigit(c))
    		{
    			digit[0] = c;
    			digit[1] = '\0';
    			digit_c = 1;
    			
    			state = 7;
    			continue;	
    		}
    		
    		fprintf(stderr,"ERROR line %d: \t '-' must be followed by a digit.\n", line);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//read digits of negative orientation gene
    	if(state == 7)
    	{
    		if(isdigit(c))
    		{
    			digit[digit_c++] = c;
    			digit[digit_c] = '\0';
    			
    			continue;	
    		}
    		
    		if(c == '$')
    		{
    			//write negative gene
    			gene = atoi(digit);

    			if(gene > gene_max)
    				gene_max = gene;
    		   			
    			//need to expand gene_mem table
    			while(gene >= gene_mem)
    			{
    				table = (int*)realloc(table, sizeof(int) * gene_mem * 2);
    				for(x = gene_mem; x < gene_mem * 2; x++)
    					table[x] = -1;
    				
    				gene_mem*=2;
    			}
    			
    			if(table[gene] == -1)
    				table[gene] = 1;
    			else
    				table[gene]++;
    		
    				        			
    			digit[0] = '\0';
    			digit_c = 0;
    			
    			//linear chromosome
    			state = 9;
    			continue;
    		}
    		
    		if(c == ';')
    		{
    			//write negative gene
    			gene = atoi(digit);
      			
    			if(gene > gene_max)
    				gene_max = gene;
    		  			
    			//need to expand gene_mem table
    			while(gene >= gene_mem)
    			{
    				table = (int*)realloc(table, sizeof(int) * gene_mem * 2);
    				for(x = gene_mem; x < gene_mem * 2; x++)
    					table[x] = -1;
    				
    				gene_mem*=2;
    			}
    			
    			if(table[gene] == -1)
    				table[gene] = 1;
    			else
    				table[gene]++;
    			
    				        			  			
    			digit[0] = '\0';
    			digit_c = 0;
    			
    			//circular chromosome	
    			state = 9;
    			continue;
    		}
    		
    		if(isspace(c))
    		{    			
    			//write negative gene
    			gene = atoi(digit);
	
    			if(gene > gene_max)
    				gene_max = gene;
    			
    			//need to expand gene_mem table
    			while(gene >= gene_mem)
    			{
    				table = (int*)realloc(table, sizeof(int) * gene_mem * 2);
    				for(x = gene_mem; x < gene_mem * 2; x++)
    					table[x] = -1;
    				
    				gene_mem*=2;
    			}
 
    			if(table[gene] == -1)
    				table[gene] = 1;
    			else
    				table[gene]++;
    			
    				        			 			
    			digit[0] = '\0';
    			digit_c = 0;
    			
    			state = 8;
    			continue;
    		}
    		fprintf(stderr,"ERROR line %d: \t Invalid character %c inside chromosome.\n", line, c);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//whitespace after a gene
    	if(state == 8)
    	{
    		if(isspace(c))
    			continue;
    		
    		if(isdigit(c))
    		{
    			digit[0] = c;
    			digit[1] = '\0';
    			digit_c++;
    			
    			state = 5;
    			continue;
    		}
    		
    		if(c == '-')
    		{
    			state = 6;
    			continue;
    		}	
    		
    		if(c == '$')
    		{
    			//linear chromosome
    			state = 9;
    			continue;
    		}
    		
    		if(c == ';')
    		{
    			//circular chromosome	
    			state = 9;
    			continue;
    		}

			fprintf(stderr,"ERROR line %d: \t Invalid character %c inside chromosome.\n", line, c);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//just finished a chromosome
    	if(state == 9)
    	{
    		if(c == EOF)
    		{
    			break;	
    		}
     		if(c == '\n')
    		{
    			line++;
    			state = 11;
    			continue;
    		}
    		if(c == '#')
    		{
    			state = 15;
    			continue;
    		}	
    		if(isspace(c))
    		{
    			state = 10;
    			continue;	
    		}
    		
    		fprintf(stderr,"ERROR line %d: \t Invalid character %c after end of chromosome.\n", line, c);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//whitespace after chromosome
    	if(state == 10)
    	{
    		if(c == EOF)
    		{
    			break;	
    		}
    		
    		if(c == '\n')
    		{
    			line++;
    			state = 11;
    			continue;
    		}

     		if(isspace(c))
    			continue;
    		
    		if(c == '#')
    		{
    			state = 15;
    			continue;
    		}	
    	}
    	
    	//comment after chromosome
    	if(state == 15)
    	{
    		if(c == EOF)
    		{
    			break;	
    		}
    		
    		if(c == '\n')
    		{
    			line++;
    			state = 11;
    			continue;	
    		}	
    		
    		continue;
    	}
    	
    	//new line after a chromosome
    	if(state == 11)
    	{
    		if(c == EOF)
    		{
    			break;	
    		}
    		
    		if(c == '#')
    		{
    			state = 15;
    			continue;	
    		}	
    		
    		if(c == '\n')
    		{
    			line++;
    			continue;	
    		}
    		
    		if(isspace(c))
    		{
    			continue;	
    		}
    		
    		if(c == '>')
    		{
    			//check for duplicates, clean out table for next pass
    			for(x = 1; x < gene_mem; x++)
    			{
    				if(table[x] > 1)
    				{
    					fprintf(stderr,"ERROR line %d: \t Genome contains multiple copies of gene %d.\n", line, x);
     					fprintf(stderr,"\tInput file requires modification.");
     					exit(-1);
    				}	
    				
    				if(table[x] == 1)
    					table[x] = 0;
    			}
    			    			
    			genome_count++;   			
    			state = 16;
    			continue;	
    		}
    		
    		if(isdigit(c))
    		{  			
    			digit[0] = c;
    			digit[1] = '\0';
    			digit_c++;
 
    			state = 5;
    			continue;
    		}
    		
    		if(c == '-')
    		{
    			state = 6;
    			continue;	
    		}
    		
    		fprintf(stderr,"ERROR line %d: \t Invalid character %c after end of chromosome.\n", line, c);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	fprintf(stderr,"ERROR state %d: \t Invalid state.\n", state);
     	exit(-1);
 
	}
	
	//check last genome for duplicates, check for missing genes
	for(x = 1; x <= gene_max; x++)
   {
   	if(table[x] > 1)
   	{
   		fprintf(stderr,"ERROR line %d: \t Genome contains multiple copies of gene %d.\n", line, x);
   		fprintf(stderr,"\tInput file requires modification.");
   		exit(-1);
   	}	
    				
   	/*if(table[x] == -1)
   	{
   		fprintf(stderr,"ERROR: \t Dataset missing gene %d.\n", x);
   		fprintf(stderr,"\tInput file requires modification.");
   		exit(-1);
   	}*/
   }
	
	rewind(fp);
	state = 1;
	line = 0;
	//allocate memory for output genomes
	(*output) = (struct genome*)malloc(sizeof(struct genome) * genome_count);
	
	for(x = 0; x < genome_count; x++)
	{
		(*output)[x].name = (char*)malloc(sizeof(char) * MAX_STR_LEN);
		(*output)[x].genes = (int*)malloc(sizeof(int) * gene_mem * 3);		
		
		for(y = 0; y < gene_max; y++)
			(*output)[x].genes[y] = 0;
	}

	//second pass
	while (1) 
  	{
		c=getc(fp);
  	
 	
    	//waiting for first genome marker
    	if(state == 1)
    	{
    		if(c == '>')
    		{
				//prepare for writing
				chr_count = 0;
				current_genome = 0;
				write_location = 1;
				current_cap = gene_max+1;
				chr_front_cap = 0;
				
				(*output)[current_genome].num_chr = 1;
				(*output)[current_genome].genes[0] = current_cap++;

    			state = 16;	
    			continue;
    		}
    		if(c == '#')
    		{
    			state = 12;
    			continue;	
    		}
    		if(c == '\n' || isspace(c))
    		{
    			continue;	
    		}

  			fprintf(stderr,"ERROR: \tExpecting '>' to mark initial genome or a comment.\n");
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//comment at beginning of file
    	if(state == 12)
    	{
    		if(c == '\n')
    		{
    			state = 1;
    			line++;
    		}
    		
    		continue;	
    	}
    	
    	//make sure name has at least one character
    	if(state == 16)
    	{
    		if(!isspace(c) && c != '\n')
    		{
   			(*output)[current_genome].name[0] = c;
   			(*output)[current_genome].name[1] = '\0';
    			name_c = 1;
    			
    			state = 2;
    			continue;
    		}
    		
    		fprintf(stderr,"ERROR line %d: \tGenome name must have at least one character.\n", line);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//reading genome name characters
    	if(state == 2)
    	{
    		if(c == '\n')
    		{
    			name_c = 0;
    			
    			line++;
    			state = 4;
    			continue;
    		}
    		
    		if(isspace(c))
    		{
    			name_c = 0;
    			
    			state = 3;
    			continue;
    		}
    		
    		if(c == '#')
    		{
    			name_c = 0;
    			
    			state = 13;
    			continue;	
    		}
    		
    		if(name_c == MAX_STR_LEN)
    		{
    			fprintf(stderr,"ERROR line %d: \tMaximum genome name size exceeded.\n", line);
     			fprintf(stderr,"\tInput file requires modification.");
     			exit(-1);
    		}
    		
    		(*output)[current_genome].name[name_c++] = c;
    		(*output)[current_genome].name[name_c] = '\0';
    		
    		continue;	
    	}
    	
    	//whitespace after genome name
    	if(state == 3)
    	{
    		if(c == '#')
    		{
    			state = 13;
    			continue;	
    		}
    		
    		if(c == '\n')
    		{
    			line++;
    			state = 4;
    			continue;	
    		}
    		
    		if(isspace(c))
    		{
    			continue;	
    		}
    		
    		fprintf(stderr,"ERROR line %d: \tWhitespace not permitted in genome name.\n", line);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//comment after genome name
    	if(state == 13)
    	{
    		if(c == '\n')
    		{
    			line++;
    			state = 4;
    			continue;	
    		}	
    		
    		continue;
    	}
    	
    	//begin reading first gene in first chromosome of genome
    	if(state == 4)
    	{
    	   if(c == '-')
    	   {
    	   	state = 6;
    	   	continue;
    	   }
    	   if(isdigit(c))
    	   {
    	   	state = 5;
    	   	
    	   	digit_c++;
    	   	digit[0] = c;
    	   	digit[1] = '\0';
    	   	
    	   	continue;
    	   }	
    	   if(c == '#')
    	   {
    	   	state = 14;
    	   	continue;	
    	   }
    	   	
    	   if(isspace(c))
    	   	continue;
    	   	
    		fprintf(stderr,"ERROR line %d: \tExpecting comment or '-' or digit to mark initial gene.\n", line);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//ignoring comment on a line after a genome name
    	if(state == 14)
    	{
    		if(c == '\n')
    		{
    			line++;
    			state = 4;
    			continue;	
    		}	
    		
    		continue;
    	}
    	
    	//read positive orientation gene
    	if(state == 5)
    	{
    		if(isdigit(c))
    		{
    			digit[digit_c++] = c;
    			digit[digit_c] = '\0';
    			
    			continue;
    		}
    		
    		if(c == '$')
    		{
    			//write gene
    			gene = atoi(digit);
    			
    			digit[0] = '\0';
    			digit_c = 0;
    			
    			(*output)[current_genome].genes[write_location++] = gene;
    			
    			//close linear chromosome
    			
    			chr_count++;
    			(*output)[current_genome].genes[write_location++] = current_cap++;
    		
    			state = 9;
    			continue;
    		}
    		
    		if(c == ';')
    		{
    			//write gene
    			gene = atoi(digit);
    				    			
    			(*output)[current_genome].genes[write_location++] = gene;
    				    			
    			digit[0] = '\0';
    			digit_c = 0;
    			
    			//close circular chromosome	
    			chr_count++;

	    		(*output)[current_genome].genes[write_location++] = current_cap * -1;
	    		current_cap++;
	    		(*output)[current_genome].genes[chr_front_cap] *= -1;	

    			state = 9;
    			continue;
    		}
    		
    		if(isspace(c))
    		{
    			//write gene
    			gene = atoi(digit);

    			(*output)[current_genome].genes[write_location++] = gene;
    				        			
    			digit[0] = '\0';
    			digit_c = 0;
    			
    			state = 8;
    			continue;
    		}
    		
    		fprintf(stderr,"ERROR line %d: \t Invalid character %c inside chromosome.\n", line, c);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//read first digit of negative orientation gene
    	if(state == 6)
    	{
    		if(isdigit(c))
    		{
    			digit[0] = c;
    			digit[1] = '\0';
    			digit_c = 1;
    			
    			state = 7;
    			continue;	
    		}
    		
    		fprintf(stderr,"ERROR line %d: \t '-' must be followed by a digit.\n", line);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//read digits of negative orientation gene
    	if(state == 7)
    	{
    		if(isdigit(c))
    		{
    			digit[digit_c++] = c;
    			digit[digit_c] = '\0';
    			
    			continue;	
    		}
    		
    		if(c == '$')
    		{
    			//write negative gene
    			gene = atoi(digit);

    			(*output)[current_genome].genes[write_location++] = gene * -1;
    				        			
    			digit[0] = '\0';
    			digit_c = 0;
    			
    			//close linear chromosome
    			
    			chr_count++;
    			(*output)[current_genome].genes[write_location++] = current_cap++;

    			state = 9;
    			continue;
    		}
    		
    		if(c == ';')
    		{
    			//write negative gene
    			gene = atoi(digit);

    			(*output)[current_genome].genes[write_location++] = gene * -1;
    				        			  			
    			digit[0] = '\0';
    			digit_c = 0;
    			
    			//close circular chromosome	
    			chr_count++;

	    		(*output)[current_genome].genes[write_location++] = current_cap * -1;
	    		current_cap++;
	    		(*output)[current_genome].genes[chr_front_cap] *= -1;	
	    		
    			state = 9;
    			continue;
    		}
    		
    		if(isspace(c))
    		{
    			//write negative gene
    			gene = atoi(digit);

    			(*output)[current_genome].genes[write_location++] = gene * -1;
   				        			 			
    			digit[0] = '\0';
    			digit_c = 0;
    			
    			state = 8;
    			continue;
    		}
    		fprintf(stderr,"ERROR line %d: \t Invalid character %c inside chromosome.\n", line, c);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//whitespace after a gene
    	if(state == 8)
    	{
    		if(isspace(c))
    			continue;
    		
    		if(isdigit(c))
    		{
    			digit[0] = c;
    			digit[1] = '\0';
    			digit_c++;
    			
    			state = 5;
    			continue;
    		}
    		
    		if(c == '-')
    		{
    			state = 6;
    			continue;
    		}	
    		
    		if(c == '$')
    		{
    			//close linear chromosome
    			chr_count++;
    			(*output)[current_genome].genes[write_location++] = current_cap++;

    			state = 9;
    			continue;
    		}
    		
    		if(c == ';')
    		{
    			//close circular chromosome	
    			
    			chr_count++;

	    		(*output)[current_genome].genes[write_location++] = current_cap * -1;
	    		current_cap++;
	    		(*output)[current_genome].genes[chr_front_cap] *= -1;	
	    		
    			state = 9;
    			continue;
    		}

			fprintf(stderr,"ERROR line %d: \t Invalid character %c inside chromosome.\n", line, c);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//just finished a chromosome
    	if(state == 9)
    	{
    		if(c == '\n')
    		{
    			line++;
    			state = 11;
    			continue;
    		}
    		if(c == '#')
    		{
    			state = 15;
    			continue;
    		}	
    		if(c == EOF)
    		{
    			//close last genome
    			(*output)[current_genome].num_marker = write_location;
    			(*output)[current_genome].num_gene = gene_max;
    			(*output)[current_genome].num_chr = chr_count;	
    		}
    		if(isspace(c))
    		{
    			state = 10;
    			continue;	
    		}

    		fprintf(stderr,"ERROR line %d: \t Invalid character %c after end of chromosome.\n", line, c);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}
    	
    	//whitespace after chromosome
    	if(state == 10)
    	{    		
    		if(c == '\n')
    		{
    			line++;
    			state = 11;
    			continue;
    		}
      	
      	if(c == EOF)
    		{
    			//close last genome
    			(*output)[current_genome].num_marker = write_location;
    			(*output)[current_genome].num_gene = gene_max;
    			(*output)[current_genome].num_chr = chr_count;	
    		}
      	
      	if(isspace(c))
    			continue;
    			  		
    		if(c == '#')
    		{
    			state = 15;
    			continue;
    		}	
    	}
    	
    	//comment after chromosome
    	if(state == 15)
    	{
    		if(c == '\n')
    		{
    			line++;
    			state = 11;
    			continue;	
    		}	
    		
    		if(c == EOF)
    		{
    			//close last chromosome
    			(*output)[current_genome].num_marker = write_location;
    			(*output)[current_genome].num_gene = gene_max;
    			(*output)[current_genome].num_chr = chr_count;	
    		
    			break;
    		}
    		
    		continue;
    	}
    	
    	//new line after a chromosome
    	if(state == 11)
    	{
    		if(c == EOF)
    		{
    			//close last chromosome
    			(*output)[current_genome].num_marker = write_location;
    			(*output)[current_genome].num_gene = gene_max;
    			(*output)[current_genome].num_chr = chr_count;	
    			break;
    		}    
    		
    		if(c == '#')
    		{
    			state = 15;
    			continue;	
    		}	
    		
    		if(c == '\n')
    		{
    			line++;
    			continue;	
    		}
    		
    		if(isspace(c))
    		{
    			continue;	
    		}
    		
    		if(c == '>')
    		{	
    			//close previous genome
    			(*output)[current_genome].num_marker = write_location;
    			(*output)[current_genome].num_gene = gene_max;
    			(*output)[current_genome].num_chr = chr_count;
    			
 				//prepare for writing
 				current_genome++;
 				chr_count = 0;
				write_location = 1;
				current_cap = gene_max+1;
				chr_front_cap = 0;
				
				(*output)[current_genome].num_chr++;
				(*output)[current_genome].genes[0] = current_cap++;
    				
    			state = 16;
    			continue;	
    		}
    		
    		if(isdigit(c))
    		{
    			//new chromosome
    			chr_front_cap = write_location;
    			(*output)[current_genome].genes[write_location++] = current_cap++;
    			
    			    			
    			digit[0] = c;
    			digit[1] = '\0';
    			digit_c = 1;
    			
    			state = 5;
    			continue;
    		}
    		
    		if(c == '-')
    		{
    			//new chromosome
    			chr_front_cap = write_location;
    			(*output)[current_genome].genes[write_location++] = current_cap++;
    			
    			state = 6;
    			continue;	
    		}		
    		
    		fprintf(stderr,"ERROR line %d: \t Invalid character %c after end of chromosome.\n", line, c);
     		fprintf(stderr,"\tInput file requires modification.");
     		exit(-1);
    	}

		fprintf(stderr,"ERROR state %d: \t Invalid state reached.\n", state);
     	exit(-1);
	}
	
	(*num_genomes) = genome_count;
}

void equalize_genome(struct genome** input, struct genome** output, int genome_count)
{
	int x, y, z;
	
	(*output) = (struct genome*)malloc(sizeof(struct genome) * genome_count);
	
	for(x = 0; x < genome_count; x++)
	{
		(*output)[x].name = (char*)malloc(sizeof(char) * MAX_STR_LEN);
		strcpy((*output)[x].name, (*input)[x].name);
		
		(*output)[x].genes = (int*)malloc(sizeof(int) * (*input)[x].num_marker);		
		
		for(y = 0; y < (*input)[x].num_marker; y++)
			(*output)[x].genes[y] = 0;
	}
	
	int *table1 = (int*)malloc(sizeof(int) * (*input)[0].num_gene);
	int *table2 = (int*)malloc(sizeof(int) * (*input)[0].num_gene);	
	
	for(x = 0; x < (*input)[0].num_gene; x++)
	{
		table1[x] = 0;
		table2[x] = 1;	
	}	
	
	int gene;
	
	for(x = 0; x < genome_count; x++)
	{
		for(y = 0; y < (*input)[x].num_marker; y++)
		{
			gene = (*input)[x].genes[y];
			
			if(ABS(gene) > (*input)[x].num_gene)
				continue;
				
			table1[ABS(gene)-1] = 1;
		}
		
		for(y = 0; y < (*input)[0].num_gene; y++)
		{
			if(!table1[y])
				table2[y] = 0;
			
			table1[y] = 0;	
		}	
	}
	
	int write, cap, next = 1, chr_status = 0;
	
	for(x = 0; x < (*input)[0].num_gene; x++)
	{		
		if(table2[x])
			table2[x] = next++;		
	}
	
	for(x = 0; x < genome_count; x++)
	{
		cap = next;
		write = 0;
		chr_status = 0;
		(*output)[x].num_gene = next-1;	

		for(y = 0; y < (*input)[x].num_marker; y++)
		{
			if(ABS((*input)[x].genes[y]) > (*input)[x].num_gene)
			{
				if(chr_status == 0 || chr_status == -1 || chr_status == 1)
				{
					if((*input)[x].genes[y] > 0)
					{
						chr_status = 1;	
					}
					else
					{
						chr_status = -1;
					}
				}
				
				if(chr_status == 2)
				{
					chr_status = 0;
					(*output)[x].genes[write++] = cap++;
				}
				
				if(chr_status == -2)
				{
					(*output)[x].genes[write++] = cap * -1;
					cap++;
					chr_status = 0;	
				}
				
				continue;
			}
			if(table2[ABS((*input)[x].genes[y])-1])
			{
				if(chr_status == 1)
				{
					(*output)[x].genes[write++] = cap++;
					chr_status = 2;	
				}
				
				if(chr_status == -1)
				{
					(*output)[x].genes[write++] = cap * -1;
					cap++;	
					chr_status = -2;
				}
				
				if((*input)[x].genes[y] > 0)
				{
					(*output)[x].genes[write++] = table2[ABS((*input)[x].genes[y])-1];
				}
				else
				{
					(*output)[x].genes[write++] = -1 * table2[ABS((*input)[x].genes[y])-1];
				}		
				continue;
			}
		}
		(*output)[x].num_marker = write;
		(*output)[x].num_chr = (cap - next + 1)/2;
	}
}

void partial_trim_genome(struct genome** input, struct genome** output, int genome_count, int threshold)
{
	int x, y, z;
	
	(*output) = (struct genome*)malloc(sizeof(struct genome) * genome_count);
	
	for(x = 0; x < genome_count; x++)
	{
		(*output)[x].name = (char*)malloc(sizeof(char) * MAX_STR_LEN);
		strcpy((*output)[x].name, (*input)[x].name);
		
		(*output)[x].genes = (int*)malloc(sizeof(int) * (*input)[x].num_marker);		
		
		for(y = 0; y < (*input)[x].num_marker; y++)
			(*output)[x].genes[y] = 0;
	}
	
	int *table1 = (int*)malloc(sizeof(int) * (*input)[0].num_gene);
	int *table2 = (int*)malloc(sizeof(int) * (*input)[0].num_gene);	
	
	for(x = 0; x < (*input)[0].num_gene; x++)
	{
		table1[x] = 0;
		table2[x] = genome_count - threshold;	
	}	
	
	int gene;
	
	for(x = 0; x < genome_count; x++)
	{
		for(y = 0; y < (*input)[x].num_marker; y++)
		{
			gene = (*input)[x].genes[y];
			
			if(ABS(gene) > (*input)[x].num_gene)
				continue;
				
			table1[ABS(gene)-1] = 1;
		}
		
		for(y = 0; y < (*input)[0].num_gene; y++)
		{
			if(!table1[y])
				table2[y]--;
			
			table1[y] = 0;	
		}	
	}
	
	int write, cap, next = 1, chr_status = 0;
	
	for(x = 0; x < (*input)[0].num_gene; x++)
	{		
		if(table2[x] > 0)
		{
			table2[x] = next++;	
		}
		else
		{
			table2[x] = 0;	
		}	
	}
	
	for(x = 0; x < genome_count; x++)
	{
		cap = next;
		write = 0;
		chr_status = 0;
		(*output)[x].num_gene = next-1;	

		for(y = 0; y < (*input)[x].num_marker; y++)
		{
			if(ABS((*input)[x].genes[y]) > (*input)[x].num_gene)
			{
				if(chr_status == 0 || chr_status == -1 || chr_status == 1)
				{
					if((*input)[x].genes[y] > 0)
					{
						chr_status = 1;	
					}
					else
					{
						chr_status = -1;
					}
				}
				
				if(chr_status == 2)
				{
					chr_status = 0;
					(*output)[x].genes[write++] = cap++;
				}
				
				if(chr_status == -2)
				{
					(*output)[x].genes[write++] = cap * -1;
					cap++;
					chr_status = 0;	
				}
				
				continue;
			}
			if(table2[ABS((*input)[x].genes[y])-1])
			{
				if(chr_status == 1)
				{
					(*output)[x].genes[write++] = cap++;
					chr_status = 2;	
				}
				
				if(chr_status == -1)
				{
					(*output)[x].genes[write++] = cap * -1;
					cap++;	
					chr_status = -2;
				}
				
				if((*input)[x].genes[y] > 0)
				{
					(*output)[x].genes[write++] = table2[ABS((*input)[x].genes[y])-1];
				}
				else
				{
					(*output)[x].genes[write++] = -1 * table2[ABS((*input)[x].genes[y])-1];
				}		
				continue;
			}
		}
		(*output)[x].num_marker = write;
		(*output)[x].num_chr = (cap - next + 1)/2;
	}
}

void jackknife_genome(struct genome** input, struct genome** output, int genome_count, int cuts)
{
	int x, y, z;
	
	(*output) = (struct genome*)malloc(sizeof(struct genome) * genome_count);
	
	for(x = 0; x < genome_count; x++)
	{
		(*output)[x].name = (char*)malloc(sizeof(char) * MAX_STR_LEN);
		strcpy((*output)[x].name, (*input)[x].name);
		
		(*output)[x].genes = (int*)malloc(sizeof(int) * (*input)[x].num_marker);		
		
		for(y = 0; y < (*input)[x].num_marker; y++)
			(*output)[x].genes[y] = 0;
	}
	
	int *table1 = (int*)malloc(sizeof(int) * (*input)[0].num_gene);
	int *table2 = (int*)malloc(sizeof(int) * (*input)[0].num_gene);	
	
	for(x = 0; x < (*input)[0].num_gene; x++)
	{
		table1[x] = 0;
		table2[x] = 1;	
	}	
	
	int gene;
	
	for(x = 0; x < cuts; x++)
	{
		int g = rand() % (*input)[0].num_gene;
		
		while(!table2[x])
			g = rand() % (*input)[0].num_gene;
		
		table2[x] = 0;
	}
	
	int write, cap, next = 1, chr_status = 0;
	
	for(x = 0; x < (*input)[0].num_gene; x++)
	{		
		if(table2[x])
			table2[x] = next++;		
	}
	
	for(x = 0; x < genome_count; x++)
	{
		cap = next;
		write = 0;
		chr_status = 0;
		(*output)[x].num_gene = next-1;	

		for(y = 0; y < (*input)[x].num_marker; y++)
		{
			if(ABS((*input)[x].genes[y]) > (*input)[x].num_gene)
			{
				if(chr_status == 0 || chr_status == -1 || chr_status == 1)
				{
					if((*input)[x].genes[y] > 0)
					{
						chr_status = 1;	
					}
					else
					{
						chr_status = -1;
					}
				}
				
				if(chr_status == 2)
				{
					chr_status = 0;
					(*output)[x].genes[write++] = cap++;
				}
				
				if(chr_status == -2)
				{
					(*output)[x].genes[write++] = cap * -1;
					cap++;
					chr_status = 0;	
				}
				
				continue;
			}
			if(table2[ABS((*input)[x].genes[y])-1])
			{
				if(chr_status == 1)
				{
					(*output)[x].genes[write++] = cap++;
					chr_status = 2;	
				}
				
				if(chr_status == -1)
				{
					(*output)[x].genes[write++] = cap * -1;
					cap++;	
					chr_status = -2;
				}
				
				if((*input)[x].genes[y] > 0)
				{
					(*output)[x].genes[write++] = table2[ABS((*input)[x].genes[y])-1];
				}
				else
				{
					(*output)[x].genes[write++] = -1 * table2[ABS((*input)[x].genes[y])-1];
				}		
				continue;
			}
		}
		(*output)[x].num_marker = write;
		(*output)[x].num_chr = (cap - next + 1)/2;
	}
}

void print_genome(struct genome* g)
{
		int x;
	
		printf("name: %s markers: %d genes: %d chromosomes: %d\n", g->name, g->num_marker, g->num_gene, g->num_chr);
		fflush(stdout);
		
		for(x = 0; x < g->num_marker; x++)
		{
			printf("%d ", g->genes[x]);
			fflush(stdout);
		}	
		printf("\n");
		fflush(stdout);
}