#ifndef STRUCTS_H
#define STRUCTS_H

#define MAX_STR_LEN 256

#define CYCLE 1
#define ODD_PATH 2
#define NEAR_EVEN 3
#define FAR_EVEN 4

#define ABS(X) (X>0?X:-X)
#define MIN(X,Y) (X<Y?X:Y)
#define MAX(X,Y) (X<Y?Y:X)

//gene into index in gene array
#define GTI(X) (X>0?(2*(X-1)):(2*ABS(X)-1))
	
//index to gene in gene array
#define ITG(X) (X%2?-((X>>1)+1):((X>>1)+1))
	
//find the other of an even/odd pair
#define TOGGLE(X) (X%2==0?X+1:X-1)

struct adj_struct
{
	//store current number of adjacencies
	//and currently allocated amount; if more needed then realloc
	int adj_count;
	int adj_mem;	
	int num_genes;
	
	//adj list length is adj_count, two adjacency slots reserved for each gene, and 2 edges on each adjacency
	//even,odd pair represents one adjacency, from 0 to 2*adj_count
	//cell contains gene number and sign of adj slot
	//num_genes * 2 + 1 is sentinel
	int *adj;
	
	//2 * num_genes, one genes for each anchor, each gene gets two cells
	//even cell is gene front (+), odd cell is gene back (-)
	//cell contains direct index into adj_list
	int *gene;
	
	//used to mark for traversals
	int *visited;
	
	//genome name
	char* name;

};

//hold information about a dcj event
//index by gene poles
//value of 0 means split to unused adjacency
struct dcj
{
	//first second
	int f, s;
	
	//the adjacency list this event occurs on
	struct adj_struct* a;
};

struct genome 
{
	char* name;
	int* genes;
	
	int num_chr;
	int num_gene;
	int num_marker;
};

struct pr
{
	int f, s;
	
	int odd_path;
	int cycle;
	
	int perp;
	int para;
	int odd;
	
	int lb;
	int ub;
};

#endif
