#include <stdlib.h>
#include <stdio.h>

#include "vertex.h"
#include "structs.h"
#include "dcj_dist.h"

struct vertex* alloc_vertex(int genomes)
{
	int x;
	
	struct vertex* out = (struct vertex*)malloc(sizeof(struct vertex));
	
	out->genome = -1;
	out->terminal = -1;
	out->active = 0;
	out->linked = NULL;
	
	out->genome_total = genomes;
	
	out->visited = 0;
	
	out->vertex_list = (struct vertex**)malloc(sizeof(struct vertex*) * out->genome_total);
	out->odd_list = (char*)malloc(sizeof(char) * out->genome_total);
	out->opposite_exit_list = (int*)malloc(sizeof(int) * out->genome_total);
	
	for(x = 0; x < out->genome_total; x++)
	{
		out->vertex_list[x] = NULL;
		out->odd_list[x] = -1;	
		out->opposite_exit_list[x] = -1;
	}
	
	return out;	
}

struct vertex_controller* initialize_vertex_controller(int initial_verticies, int genomes)
{
	int x;
	struct vertex_controller* out = (struct vertex_controller*)malloc(sizeof(struct vertex_controller));
	
	out->initial_edges = genomes;
	
	out->vertex_total = initial_verticies;
	out->vertex_count = 0;
	
	out->vertex_memory = (struct vertex**)malloc(sizeof(struct vertex*) * out->vertex_total);
	
	for(x = 0; x < out->vertex_total; x++)
	{
		out->vertex_memory[x] = alloc_vertex(out->initial_edges);	
	}
	
	return out;	
}

struct vertex* get_vertex(struct vertex_controller* vc)
{
	if(vc->vertex_count+1 >= vc->vertex_total)
	{
		int x;
		
		vc->vertex_memory = (struct vertex**)realloc(vc->vertex_memory, sizeof(struct vertex*) * vc->vertex_total * 2);
	
		for(x = vc->vertex_total; x < vc->vertex_total * 2; x++)
		{
			vc->vertex_memory[x] = alloc_vertex(vc->initial_edges);	
		}
		
		vc->vertex_total *= 2;
	}
	
	return vc->vertex_memory[vc->vertex_count++];
}

void return_vertex(struct vertex_controller* vc, struct vertex* v)
{
	int x, y;
	
	for(x = 0; x < vc->vertex_count; x++)
	{
		if(vc->vertex_memory[x] == v)
		{
			vc->vertex_memory[x] = vc->vertex_memory[--vc->vertex_count];	

			v->genome = -1;
			v->terminal = 0;
			v->visited = 0;
			v->active = 0;
			
			for(y = 0; y < v->genome_total; y++)
			{
				v->vertex_list[y] = NULL;
				v->odd_list[y] = -1;	
				v->opposite_exit_list[x] = -1;
			}

			break;
		}	
	}
}

void print_vertex(struct vertex* v)
{
	int x;
	
	if(v == NULL)
	{
		printf("empty vertex\n");
		return;	
	}
	
	printf("self: %d genome: %d gene terminal: %d active: %d\n", v, v->genome, v->terminal, v->active);
	printf("edge memory: %d\nedges:\n", v->genome_total);	
	for(x = 0; x < v->genome_total; x++)
	{
		if(v->vertex_list[x])
		{
			printf("\t to vertex: %d genome: %d terminal: %d odd? %d opposite exit: %d\n", v->vertex_list[x], v->vertex_list[x]->genome, v->vertex_list[x]->terminal, v->odd_list[x], v->opposite_exit_list[x]);
		}
		else
		{
			printf("\t empty\n");	
		}
	}
	printf("\n");
}

void clear_visits(struct vertex_controller* vc)
{
	int x;
	
	for(x = 0; x < vc->vertex_count; x++)
	{
		vc->vertex_memory[x]->visited = 0;	
	}	
}

struct vertex* make_vertex(int genome, int terminal, struct vertex_controller* vc)
{
	struct vertex* out = (struct vertex*)get_vertex(vc);
	
	out->genome = genome;
	out->terminal = terminal;
	out->active = 1;
	
	return out;
}

void add_path(struct vertex* v, int exit_genome, char odd)
{
	v->vertex_list[exit_genome] = NULL;
	v->odd_list[exit_genome] = odd;
}

void add_parallel(struct vertex* v1, int exit1, struct vertex* v2, int exit2)
{
	v1->vertex_list[exit1] = v2;
	v2->vertex_list[exit2] = v1;
	
	v1->odd_list[exit1] = 1;
	v2->odd_list[exit2] = 1;
	
	v1->opposite_exit_list[exit1] = exit2;
	v2->opposite_exit_list[exit2] = exit1;
}

//assume they are both in the same genome, even will be set as 1
void add_perpendicular(struct vertex* v1, struct vertex* v2, int exit)
{	
	v1->vertex_list[exit] = v2;
	v2->vertex_list[exit] = v1;
		
	v1->odd_list[exit] = 0;
	v2->odd_list[exit] = 0;
	
	v1->opposite_exit_list[exit] = exit;
	v2->opposite_exit_list[exit] = exit;
}

struct vertex* find_vertex(int genome, int terminal, struct vertex_controller* vc)
{
	int x;
	
	for(x = 0; x < vc->vertex_count; x++)
	{
		if(genome == vc->vertex_memory[x]->genome && terminal == vc->vertex_memory[x]->terminal)
		{
			return vc->vertex_memory[x];	
		}
	}
	
	return NULL;
}

void build_graph(struct adj_struct** in, int genomes, struct vertex_controller* vc)
{
	struct adj_struct* a;
	struct adj_struct* b;
	
	int i, j, x, t, next;

	for(i = 0; i < genomes; i++)
	{
		for(j = 0; j < in[i]->num_genes * 2; j++)
		{
			if(in[i]->gene[j] == -1)
			{
				make_vertex(i, ITG(j), vc);
			}
		}	
	}

	for(i = 0; i < genomes-1; i++)
	{
		for(j = i+1; j < genomes; j++)
		{
			//printf("i:%d j:%d\n", i, j);
			//fflush(stdout);
			
			a = in[i];
			b = in[j];
			
			for(x = 0; x < a->num_genes*2; x++)
			{
				a->visited[x] = 0;
				b->visited[x] = 0;
			}
			
			//find our prosthetic chromosome components
			//for each genome find missing genes and begin
			//a traversal from each,  mark visited adjacencies 
			//but don't skip them; each perp component will
			//be visited twice so divide by two
			
			//genome a
			for(x = 0; x <= a->num_genes*2; x++)
			{		
				if(a->gene[x] == -1)
				{		
					t = x;
					next = b->gene[x];
					
					while(1)
					{
						//parallel component
						if(next == -1)
						{
							//printf("parallel a (%d,%d,%d) (%d,%d,%d)\n", i, ITG(x), j, j, ITG(t), i);
							//fflush(stdout);
							add_parallel(find_vertex(i, ITG(x),vc),j,find_vertex(j, ITG(t), vc),i);
							break;	
						}	
		
						b->visited[next>>1]++;				
						next = b->adj[TOGGLE(next)];
						
						//odd component
						if(next == a->num_genes*2+1)
						{
							//printf("odd component (%d,%d,%d)\n",i, ITG(x), 1);
							//fflush(stdout);
							add_path(find_vertex(i, ITG(x), vc),j,1);
							break;
						}
						
						t = next;
						next = a->gene[next];
						
						//perpendicular component
						if(next == -1)
						{
							//printf("perpendicular (%d,%d,%d) (%d,%d,%d)\n", i, ITG(x), j, i, ITG(t), j);
							//fflush(stdout);
							add_perpendicular(find_vertex(i, ITG(x), vc),find_vertex(i, ITG(t), vc), j);
							break;	
						}
		
						a->visited[next>>1]++;				
						next = a->adj[TOGGLE(next)];	
						
						//even component
						if(next == a->num_genes*2+1)
						{
							//printf("even component (%d,%d,%d)\n",i, ITG(x), 0);
							//fflush(stdout);
							add_path(find_vertex(i, ITG(x), vc),j,0);
							break;
						}
						
						t = next;
						next = b->gene[next];
					}
				}
			}
			
			//genome b
			for(x = 0; x <= b->num_genes*2; x++)
			{
				if(b->gene[x] == -1)
				{	
					t = x;
					next = a->gene[x];
					
					/*if(j == 3 && i == 0)
					{
						printf("found the right search j%d i%d x%d next%d\n", j, i, x, next);
						adj_print(in[i]);
						adj_print(in[j]);
					}*/
					while(1)
					{
		
						//parallel component
						if(next == -1)
						{
							//printf("parallel b (%d,%d,%d) (%d,%d,%d)\n", j, ITG(x), i, i, ITG(t), j);
							//fflush(stdout);
							add_parallel(find_vertex(j, ITG(x), vc), i, find_vertex(i, ITG(t), vc), j);							

							break;	
						}	
		
						a->visited[next>>1]++;				
						next = a->adj[TOGGLE(next)];
				
						//if(j == 3 && i == 0)
							//printf("next%d\n", next);
					
				
						//odd component
						if(next == b->num_genes*2+1)
						{
							//printf("odd component (%d,%d,%d)\n",j, ITG(x), 1);
							//fflush(stdout);
							add_path(find_vertex(j, ITG(x), vc),i,1);
							break;
						}
						
						t = next;
						next = b->gene[next];
						
						//if(j == 3 && i == 0)
							//printf("next%d\n", next);
						
						//perpendicular component
						//again, do the single vertex per visit
						if(next == -1)
						{	
							//printf("perpendicular (%d,%d,%d) (%d,%d,%d)  %d=%d\n", j, ITG(x), i, j, ITG(t), i, x, t);
							//fflush(stdout);
							add_perpendicular(find_vertex(j, ITG(x), vc), find_vertex(j, ITG(t), vc), i);	
							break;	
						}
		
						b->visited[next>>1]++;				
						next = b->adj[TOGGLE(next)];
		
						//if(j == 3 && i == 0)
							//printf("next%d\n", next);
		
						//even component
						if(next == b->num_genes*2+1)
						{
						//	printf("even component (%d,%d,%d)\n",j, ITG(x), 0);
						//	fflush(stdout);
							add_path(find_vertex(j, ITG(x), vc), i, 0);
							break;
						}
						
						t = next;
						next = a->gene[next];
						
						//if(j == 3 && i == 0)
							//printf("next%d\n", next);
					}
				}
			}
		}
	}	
}

void bound_graph(struct vertex_controller* vc, int* upper_bound, int* lower_bound)
{
	int even = 0, odd = 0, perp = 0, para = 0;
	
	int own_genome;
	struct vertex* target_vertex;
	int target_odd;
	
	int x, y;
	
	for(x = 0; x < vc->vertex_count; x++)
	{
		if(!vc->vertex_memory[x]->active)
			continue;
		
		own_genome = vc->vertex_memory[x]->genome;
		
		for(y = 0; y < vc->vertex_memory[x]->genome_total; y++)
		{	
			target_vertex = vc->vertex_memory[x]->vertex_list[y];
			target_odd = vc->vertex_memory[x]->odd_list[y];
			
			if(target_odd == -1)
				continue;
			
			//even component
			if(target_vertex == 0 && target_odd == 0)
			{
				even++;
				continue;
			}
			
			//odd component
			if(target_vertex == 0 && target_odd == 1)
			{
				odd++;
				continue;
			}
			
			//parallel component
			if(own_genome != target_vertex->genome)
			{
				para++;
				continue;
			}
			
			//perpendicular component
			if(own_genome == target_vertex->genome)
			{
				perp++;
				continue;
			}	
			
		}
	}
		
	(*upper_bound) = odd + para + 2 * perp;
	(*lower_bound) = odd + para;
}

void cap_vertex(struct vertex* v, struct graph_modifier* gm)
{
	int x, y;
	struct vertex* target;
	gm->double_gain = 0;
	gm->v1 = v;
	gm->v2 = NULL;
	
	v->active = 0;
	
	for(x = 0; x < v->genome_total; x++)
	{
		if(v->odd_list[x] == -1)
			continue;
		
		if(v->vertex_list[x] == 0)
		{
			if(v->odd_list[x] == 0)
			{
				gm->double_gain++;	
			}
			continue;	
		}
		
		target = v->vertex_list[x];
		target->vertex_list[v->opposite_exit_list[x]] = NULL;
	}
}

void merge_vertex(struct vertex* v1, struct vertex* v2, struct graph_modifier* gm)
{
	int x, y, i, j;
	struct vertex *v3, *v4;
	
	v1->active = 0;
	v2->active = 0;
	
	gm->double_gain = 0;
	gm->v1 = v1;
	gm->v2 = v2;
	
	v1->linked = v2;
	v2->linked = v1;
	
	for(x = 0; x < v1->genome_total; x++)
	{
		v3 = v1->vertex_list[x];
		v4 = v2->vertex_list[x];
		
		if(v3 == 0 && v4 == 0)
			continue;
		
		if(v3 == 0)
		{
			j = v2->opposite_exit_list[x];
			
			v4->vertex_list[j] = 0;
			v4->odd_list[j] = (v4->odd_list[j] +	v2->odd_list[x]) % 2;
			
			continue;
		}
		
		if(v4 == 0)
		{
			i = v1->opposite_exit_list[x];
			
			v3->vertex_list[i] = 0;
			v3->odd_list[i] = (v3->odd_list[i] + v1->odd_list[x]) % 2;
			
			continue;
		}
		
		if(v1 == v4)
		{
			gm->double_gain += 2;
			continue;	
		}
		
		i = v1->opposite_exit_list[x];			
		j = v2->opposite_exit_list[x];
		
		v3->odd_list[i] = (v3->odd_list[i] + v4->odd_list[j]) % 2; 
		v4->odd_list[j] = v3->odd_list[i];
		
		v3->vertex_list[i] = v4;
		v4->vertex_list[j] = v3;
		
		v3->opposite_exit_list[i] = j;
		v4->opposite_exit_list[j] = i;
	}
}

void replace_vertex(struct graph_modifier* gm)
{
	gm->v1->linked = NULL;
	gm->v2->linked = NULL;
	int x;
	struct vertex* target;
	
	
	
	if(gm->v2 == NULL)
	{
		gm->v1->active = 1;
		
		for(x = 0; x < gm->v1->genome_total; x++)
		{
			if(gm->v1->vertex_list[x])
			{
				target = gm->v1->vertex_list[x];
				
				target->vertex_list[gm->v1->opposite_exit_list[x]] = gm->v1;
				target->opposite_exit_list[gm->v1->opposite_exit_list[x]] = x;
				target->odd_list[gm->v1->opposite_exit_list[x]] = gm->v1->odd_list[x];
			}
		}
	}
	else
	{
		gm->v1->active = 1;
		gm->v2->active = 1;
		
		for(x = 0; x < gm->v1->genome_total; x++)
		{
			if(gm->v1->vertex_list[x])
			{
				target = gm->v1->vertex_list[x];
				
				target->vertex_list[gm->v1->opposite_exit_list[x]] = gm->v1;
				target->opposite_exit_list[gm->v1->opposite_exit_list[x]] = x;
				target->odd_list[gm->v1->opposite_exit_list[x]] = gm->v1->odd_list[x];
			}
			
			if(gm->v2->vertex_list[x])
			{
				target = gm->v2->vertex_list[x];
				
				target->vertex_list[gm->v2->opposite_exit_list[x]] = gm->v2;
				target->opposite_exit_list[gm->v2->opposite_exit_list[x]] = x;
				target->odd_list[gm->v2->opposite_exit_list[x]] = gm->v2->odd_list[x];
			}
		}
	}
	
}

void print_graph_modifier(struct graph_modifier* gm)
{
	print_vertex(gm->v1);
	print_vertex(gm->v2);
	
	printf("double gain: %d\n", gm->double_gain);
}

int judge_cap(struct vertex* v)
{
	int x;
	int gain = 0;
	int loss = 0;
	
	for(x = 0; x < v->genome_total; x++)
	{
		if(v->odd_list[x] == -1)
			continue;
			
		if(v->vertex_list[x] == 0 && v->odd_list[x])
		{
			gain++;
			continue;	
		}
		
		if(v->vertex_list[x]->genome != v->genome && !v->odd_list[x])
		{
			loss++;
			continue;	
		}
		
		if(v->vertex_list[x]->genome != v->genome && v->odd_list[x])
		{
			loss++;
			continue;
		}
		
		if(v->vertex_list[x]->genome == v->genome && !v->odd_list[x])
		{
			loss+=2;
			continue;	
		}
		
		if(v->vertex_list[x]->genome == v->genome && v->odd_list[x])
		{
			loss+=2;
			continue;	
		}
	}
	
			
	if(gain >= loss)
	{
		return 1;	
	}
		
	return 0;	
}

int judge_splice(struct vertex* v1, struct vertex* v2)
{
	int x;
	
	int gain = 0, loss = 0, risk = 0;
	
	for(x = 0; x < v1->genome_total; x++)
	{
		//printf("x: %d\n");
		
		if(x == v1->genome)
			continue;
		
		//perpendicular component
		if(v1->vertex_list[x] == v2)
		{
			//printf("+2\n");
			gain+=2;
			continue;	
		}
		
		//parallel components both odd
		if(v1->vertex_list[x] && v2->vertex_list[x] && v1->odd_list[x] && v2->odd_list[x])
		{

			if(v1->vertex_list[x]->genome == v2->vertex_list[x]->genome)
			{
				//printf(",2");
				risk+=2;
			}
			//printf(")\n");
			
			continue;
		}
		
		//odd components
		if(v1->vertex_list[x] == 0 && v2->vertex_list[x] == 0 && v1->odd_list[x] && v2->odd_list[x])
		{
			//printf("(-2)\n");
			loss+=2;	
			continue;
		}
		
		//one odd one parallel odd
		if(v1->vertex_list[x] == 0 && v2->vertex_list[x] && v1->odd_list[x] && v2->odd_list[x])
		{
			//printf("(-2)\n");
			loss+=2;	
			continue;
		}
		
		//other one odd, one parallel odd
		if(v1->vertex_list[x] && v2->vertex_list[x] == 0 && v1->odd_list[x] && v2->odd_list[x])
		{
			//printf("(-2)\n");
			loss+=2;	
			continue;
		}
	}

	printf("gain: %d loss: %d risk: %d\n", gain, loss, risk);

	if(gain == 0)
		return -1;

	if(gain > loss + risk)
		return 1;
	
	if(gain > loss)
		return 0;
		
	return -1;
}

int gain(struct vertex* v1, struct vertex* v2)
{
	int x;
	
	int gain = 0;
	
	for(x = 0; x < v1->genome_total; x++)
	{
		if(x == v1->genome)
			continue;
		
		if(!v1->active || !v2->active)
			continue;
		
		//perpendicular component
		if(v1->vertex_list[x] == v2)
		{
			gain+=2;
			continue;	
		}
	}
	
	return gain;
}

int risk(struct vertex *v1, struct vertex *v2)
{
	int x;
	
	int risk = 0;
	
	for(x = 0; x < v1->genome_total; x++)
	{
		if(x == v1->genome)
			continue;
		
		if(!v1->active || !v2->active)
			continue;
	
		//parallel components both odd
		if(v1->vertex_list[x] && v2->vertex_list[x] && v1->odd_list[x] == 1 && v2->odd_list[x] == 1)
		{
			if(v1->vertex_list[x]->genome == v2->vertex_list[x]->genome)
			{
				risk+=2;
			}			
			continue;
		}	
	}
	
	return risk;
}

int loss(struct vertex *v1, struct vertex *v2)
{
	int x;
	
	int loss = 0;
	
	for(x = 0; x < v1->genome_total; x++)
	{

		if(x == v1->genome)
			continue;
		
		if(!v1->active || !v2->active )
			continue;

		
		//odd components
		if(v1->vertex_list[x] == 0 && v2->vertex_list[x] == 0 && v1->odd_list[x] == 1 && v2->odd_list[x] == 1)
		{
			//printf("mark1\n");
			loss+=2;	
			continue;
		}
		
		//one odd one parallel odd
		if(v1->vertex_list[x] == 0 && v2->vertex_list[x] && v1->odd_list[x] == 1 && v2->odd_list[x] == 1)
		{
			//printf("mark2\n");
			loss+=2;	
			continue;
		}
		
		//other one odd, one parallel odd
		if(v1->vertex_list[x] && v2->vertex_list[x] == 0 && v1->odd_list[x] == 1 && v2->odd_list[x] == 1)
		{
			//printf("mark3\n");
			loss+=2;	
			continue;
		}

		//parallel to different genomes
		if(v1->vertex_list[x] && v2->vertex_list[x] && v1->vertex_list[x]->genome != v2->vertex_list[x]->genome && v1->odd_list[x] == 1 && v2->odd_list[x] == 1)
		{
			//printf("mark4\n");
			loss+=2;
			continue;	
		}
	}
	
	return loss;
}

void search_v1(struct vertex_controller* vc, int gain)
{
	int x, y;
	
	static int best_gain = 0;
	int ub, lb;
	
	struct graph_modifier gm;
	struct vertex *v1, *v2;
	
	bound_graph(vc, &ub, &lb);
	
	if(gain > best_gain)
	{
		printf("new best, gain: %d \n", gain);
		fflush(stdout);
		best_gain = gain;	
	}
	
	if(lb + gain >= best_gain)
	{
		for(x = 0; x < vc->vertex_count-1; x++)
		{	
			v1 = vc->vertex_memory[x];

			//printf("vertex x:\n");
			//print_vertex(v1);

			if(!v1->active)
				continue;
			
			for(y = x; y < v1->genome_total; y++)
			{
				v2 = v1->vertex_list[y];
				
				//printf("vertex y:\n");
				//print_vertex(v2);
				
				if(v2 == 0 || !v2->active)
					continue;
				
				if(has_perp(v1, v2))
				{
					merge_vertex(v1, v2, &gm);
					
					printf("merge v1: %d,%d v2: %d,%d\n", gm.v1->genome, gm.v1->terminal, gm.v2->genome, gm.v2->terminal);
					
					search_v1(vc, gain + gm.double_gain);
					replace_vertex(&gm);
				}
			}
			
			if(v1->active)
			{
				cap_vertex(v1, &gm);
				search_v1(vc, gain + gm.double_gain);
				replace_vertex(&gm);
			}
		}
	}
}

int has_perp(struct vertex* v1, struct vertex* v2)
{
	int x;
	
	if(v1->genome != v2->genome)
		return 0;
	
	for(x = 0; x < v1->genome_total; x++)
	{
		if(v1->vertex_list[x] == v2)
			return 1;
	}
	
	return 0;
}

int find_optimal(struct vertex_controller* vc)
{
	int x, y, judgement;
	struct vertex *v1, *v2;
	struct graph_modifier gm;	
	int ub, lb;
	bound_graph(vc, &ub, &lb);
	printf("current ub: %d lb: %d\n", ub, lb);
		
	for(x = 0; x < vc->vertex_count-1; x++)
	{
		v1 = vc->vertex_memory[x];
		if(!v1->active)
			continue;
		
		for(y = 0; y < v1->genome_total; y++)
		{
			v2 = v1->vertex_list[y];
			
			if(!v2 || !v2->active || !v1->odd_list[y] == -1)
				continue;
				
			if(judge_splice(v1, v2) >= 0)
			{
				merge_vertex(v1, v2, &gm);
				printf("merging vertex:\n");
				print_vertex(v1);
				print_vertex(v2);
				return 1;
			}	
		}	
	}
	
	return 0;
}

int measure_uncertainty(struct vertex** vertex_list, int list_size)
{
	int x, y, z;
	int flag, uncertainty = 0;
	struct vertex *v1, *v2;
	
	for(x = 0; x < list_size; x++)
	{
		v1 = vertex_list[x];
		
		for(y = 0; y < v1->genome_total; y++)
		{
			v2 = v1->vertex_list[y];
			
			if(v2 == 0 || !v2->active || v1->odd_list[y] == -1)
				continue;
			
			for(z = 0; z < list_size; z++)
			{	
				if(vertex_list[z] == v2)
				{
					break;
				}
			}
			
			if(z == list_size)
				uncertainty++;
		}
	}
	
	return uncertainty;
}

int measure_gain(struct vertex** vertex_list, int list_size)
{
	struct graph_modifier gm;
	
	int x;
	
	for(x = 2; x <= list_size; x+=2)
	{
		measure_gain_recursion(0, x, vertex_list, list_size);
	}
	
	return measure_gain_recursion(0,-1, NULL, 0);
}

int measure_gain_recursion(int gain, int depth, struct vertex** vertex_list, int list_size)
{	
	int x, y;
	struct vertex *v1, *v2;
	struct graph_modifier gm;
	
	static int best = 0;
	
	//reset best
	if(depth == -1)
	{
	int t = best;
		best = 0;
		return t;	
	}
	
	if(depth == 0)
	{
		if(gain > best)
			best = gain;
			
		return 0;
	}	
	for(x = 0; x < list_size-1; x++)
	{

		v1 = vertex_list[x];
		
		if(v1->visited || v1 == 0 || !v1->active)
			continue;
		
		v1->visited = 1;
		for(y = x+1; y < list_size; y++)
		{
			v2 = vertex_list[y];
			
			if(v2->visited || v2 == 0 || !v2->active)
				continue;
			
			v2->visited = 1;
			if(has_perp(v1, v2))
			{				
				merge_vertex(v1, v2, &gm);				
				measure_gain_recursion(gain + gm.double_gain, depth-2, vertex_list, list_size);		
				replace_vertex(&gm);
			}
			v2->visited = 0;
		}
		v1->visited = 0;
	}
	
	return 0;
}

void search_v2(struct adj_struct** adj_list, int num_genomes)
{
	struct vertex_controller* vc = (struct vertex_controller*)initialize_vertex_controller(10, num_genomes);
	build_graph(adj_list, num_genomes, vc);
	

	int x, y;
	
	int total_gain = 0;
	
	int g, l, r;
	
	int lb, ub;
	
	struct vertex *v1, *v2;
	struct graph_modifier gm;
	
	double best_gain;
	int best_genome;
	struct vertex *best_v1, *best_v2;
	
	int flag = 1;
	
	while(flag)
	{
		flag = 0;
		best_gain = 0.0;
	
		for(x = 0; x < vc->vertex_count - 1; x++)
		{
			v1 = vc->vertex_memory[x];
			
			if(!v1->active)
				continue;
			
			for(y = x+1; y < vc->vertex_count; y++)
			{
				v2 = vc->vertex_memory[y];
				
				if(!v1->active)
					break;
				
				if(!v2->active)
					continue;
				
				if(has_perp(v1, v2))
				{
					int vertex_count = 1;

					g = search_perp_v2(v1, v2, vc, &vertex_count);
					
					if((double)g/(double)vertex_count > best_gain)
					{
						best_gain = (double)g/(double)vertex_count;
						best_genome = v1->genome;
						best_v1 = v1;
						best_v2 = v2;
					}
				}
			}
		}
		
		if(best_gain != 0)
		{		
			merge_vertex(best_v1, best_v2, &gm);
			total_gain+=gm.double_gain;
			//bound_graph(vc, &ub, &lb);

			add_adj(adj_list[best_genome], best_v1->terminal, best_v2->terminal);
			flag = 1;
		}
	}
	
	
	for(x = 0; x < num_genomes; x++)
	{
		for(y = 0; y < adj_list[x]->num_genes*2; y++)
		{
			if(adj_list[x]->gene[y] == -1)
			{
				add_adj(adj_list[x], ITG(y), 0);	
			}	
		}
	}
}

void search_v2fast(struct adj_struct** adj_list, int num_genomes)
{
	struct vertex_controller* vc = (struct vertex_controller*)initialize_vertex_controller(10, num_genomes);
	build_graph(adj_list, num_genomes, vc);

	int x, y;
	
	int total_gain = 0;
	
	int g, l, r;
	
	int lb, ub;
	
	
	struct vertex *v1, *v2;
	struct graph_modifier gm;
	
	int flag = 1;
	
	while(flag)
	{
		flag = 0;

		for(x = 0; x < vc->vertex_count - 1; x++)
		{
			v1 = vc->vertex_memory[x];
			
			if(!v1->active)
				continue;
			
			for(y = x+1; y < vc->vertex_count; y++)
			{
				v2 = vc->vertex_memory[y];
				
				if(!v1->active)
					break;
				
				if(!v2->active)
					continue;
				
				if(has_perp(v1, v2))
				{
					int blah;
					g = search_perp_v2(v1, v2, vc, &blah);
					
					//printf("gain: %d\n", g);
					//fflush(stdout);
					
					if(g > 0)
					{
						merge_vertex(v1, v2, &gm);
						total_gain+=gm.double_gain;
						//bound_graph(vc, &ub, &lb);
									
						//fflush(stdout);
						add_adj(adj_list[v1->genome], v1->terminal, v2->terminal);
						flag = 1;
					}
				}
			}
		}
	}
	
	
	for(x = 0; x < num_genomes; x++)
	{
		for(y = 0; y < adj_list[x]->num_genes*2; y++)
		{
			if(adj_list[x]->gene[y] == -1)
			{
				add_adj(adj_list[x], ITG(y), 0);	
			}	
		}
	}
}

int search_perp(struct vertex* v1, struct vertex* v2)
{
	int x, single_gain;
	
	int g = gain(v1,v2);
	int l = loss(v1,v2);
	int r = risk(v1,v2);

	struct vertex *v3, *v4;

	int have = g - l - r;
	
	struct graph_modifier gm;
	
	if(v1 == v2)
		printf("introduce\n");
	
	printf("perp search: (%d,%d) (%d,%d)\n", v1->genome, v1->terminal, v2->genome, v2->terminal);
	printf("gain: %d risk: %d loss: %d\n", g, r, l);
	fflush(stdout);
	
	print_vertex(v1);
	print_vertex(v2);
	
	merge_vertex(v1, v2, &gm);
	
	for(x = 0; x < v1->genome_total; x++)
	{
		printf("x: %d\n", x);
			
		if(!v1->vertex_list[x] || !v2->vertex_list[x])
			continue;	

		printf("step1\n");
		
		v3 = v1->vertex_list[x];
		v4 = v2->vertex_list[x];
	
		if(!v3->active || !v4->active)
			continue;
	
		printf("step2\n");
	
		if(x == v1->genome)
			continue;
		
		printf("search: gs: (%d,%d)\n", v3->genome, v4->genome);
		
		if(v3->genome == v4->genome)
		{
			printf("search more\n");
			
			//single_gain = search_perp(v3, v4);
			
			printf("x: %d single gain: %d\n", x, single_gain);
			
			if(single_gain >= 2)
				have += single_gain;	
		}
	}
	
	replace_vertex(&gm);
	
	return have;	
}

int search_perp_watch(struct vertex* v1, struct vertex* v2, struct vertex* w1, struct vertex* w2, int depth)
{	
	int x, y, single_gain;
	
	int g = gain(v1,v2);
	int l = loss(v1,v2);
	int r = risk(v1,v2);

	struct vertex *v3, *v4;

	int have = g - l - r;
	
	struct graph_modifier gm;
	
	int pruned = 0, searched = 0;
	
	//printf("depth: %d\n", depth);
	
	//printf("perp search: (%d,%d) (%d,%d)\n", v1->genome, v1->terminal, v2->genome, v2->terminal);
	//printf("gain: %d risk: %d loss: %d\n", g, r, l);
	//fflush(stdout);
	
	//printf("status of vertex pair: %d\n", check_vertex_pair(v1,v2));

	/*print_vertex(v1);
	print_vertex(v2);
	print_vertex(w1);
	print_vertex(w2);*/
	//print_vertex(w3);
	//print_vertex(w4);
	//print_vertex(w5);
	
	if(!check_opposite_exit(w1) || !check_opposite_exit(w2))
	{
		printf("failed3\n");
		exit(0);
	}
	
	if(!check_active_link(w1) || !check_active_link(w2))
	{
		printf("failed2\n");
		exit(0);	
	}
	
	if(!check_vertex_pair(w1,w2) && w1->active && w2->active)
	{
		printf("failed\n");
		exit(0);
	}

	merge_vertex(v1, v2, &gm);
	
	for(x = 0; x < v1->genome_total; x++)
	{
	//	printf("x: %d\n", x);
			
		v3 = v1->vertex_list[x];
		v4 = v2->vertex_list[x];
		
		if(!v3 || !v4)
			continue;	
	
		if(!v3->active || !v4->active)
			continue;
	
		//printf("step2\n");
	
		if(x == v1->genome)
			continue;
		
		//printf("search: gs: (%d,%d)\n", v3->genome, v4->genome);
		//fflush(stdout);
		//
		if(v3->genome != v4->genome)
			continue;
	
		int flag = 1;
		for(y = x+1; y < v1->genome_total; y++)
		{
			struct vertex* y1 = v1->vertex_list[y];
			struct vertex* y2 = v2->vertex_list[y];
			
			if(!y1 || !y2)
				continue;
			if(y == v1->genome)
				continue;
			if(y1->genome != y2->genome)
				continue;
				
			//printf("cycle search: %d %d\n", x, y);
				//fflush(stdout);
			
			if(cycle_linked(v3,v4,y1,y2,v1,v2))
			{
				//printf("cycle found: %d %d\n", x, y);
				//fflush(stdout);
				flag = 0; 
				break;
			}
		}
	
		if(!flag)
		{
			pruned++;
			//printf("flag wrong?\n");
			//fflush(stdout);
			continue;
		}
		searched++;
		single_gain = search_perp_watch(v3, v4, w1, w2, depth+1);
			
		//printf("x: %d single gain: %d\n", x, single_gain);
		//fflush(stdout);
			
		if(single_gain >= 2)
			have += single_gain;	
	}
	
	replace_vertex(&gm);
	
	
	int fart1 = gm.v1->genome;
	int fart2 = gm.v1->terminal;
	int fart3 = gm.v2->genome;
	int fart4 = gm.v2->terminal;
	
	//printf("after replacment: (%d,%d) (%d,%d)\n", fart1, fart2, fart3, fart4);
	/*print_vertex(v1);
	print_vertex(v2);
	print_vertex(w1);
	print_vertex(w2);*/
	//print_vertex(w3);
	//print_vertex(w4);
	//print_vertex(w5);
	
	if(!check_opposite_exit(w1) || !check_opposite_exit(w2))
	{
		printf("failed3\n");
		exit(0);
	}
	
	if(!check_active_link(w1) || !check_active_link(w2))
	{
		printf("failed2\n");
		exit(0);	
	}
	
	if(!check_vertex_pair(w1,w2) && w1->active && w2->active)
	{
		printf("failed\n");
		exit(0);
	}
	
	if(have > 400000)
	{
		printf("have %d at depth %d pruned %d searched %d\n", have, depth, pruned, searched);
		fflush(stdout);
	}
	if(have < 0)
	{
		printf("overflowed\n");
		exit(0);	
	}
	return have;	
}

int check_vertex_pair(struct vertex *v1, struct vertex *v2)
{
	int count = 0;
	int x, y;

	for(x = 0; x < v1->genome_total; x++)
	{
		struct vertex *t = v1->vertex_list[x];
		
		if(v2 == t)
			count++; 
	}
		
	for(x = 0; x < v1->genome_total; x++)
	{
		struct vertex *t = v2->vertex_list[x];
		
		if(v1 == t)
			count--;	

	}

	return !count;
}

int check_active_link(struct vertex *v)
{
	int x;
	
	for(x = 0; x < v->genome_total; x++)
	{
		struct vertex *t = v->vertex_list[x];
		
		if(v->active && t && !t->active)
		{
			printf("(%d, %d) (%d, %d)\n", v->genome, v->terminal, t->genome, t->terminal);
			return 0;
		}
	}
	
	return 1;		
}

int check_opposite_exit(struct vertex *v)
{
	int x, y;
	
	for(x = 0; x < v->genome_total-1; x++)
	{
		struct vertex *v1 = v->vertex_list[x];
		
		if(!v1)
			continue;
		
		for(y = x+1; y < v->genome_total; y++)
		{
			struct vertex *v2 = v->vertex_list[y];
			
			if(!v2)
				continue;
			
			if(v1->active && v2->active && v1->genome == v2->genome && v1->terminal == v2->terminal && v->opposite_exit_list[x] == v->opposite_exit_list[y])
			{
				printf("shared opposites?\n");
				print_vertex(v);
				print_vertex(v1);
				print_vertex(v2);
				return 0;	
			}
		}	
	}
	
	return 1;	
}

int cycle_linked(struct vertex* v1, struct vertex* v2, struct vertex* v3, struct vertex* v4, struct vertex *w1, struct vertex *w2)
{	
	int x;

	w1->visited = 1;
	w2->visited = 1;
	
	x = no_cycle_search(v1,v2,v3,v4);
	
	clear_visited(w1);
	clear_visited(w2);

	return x;	
}

int no_cycle_search(struct vertex* v1, struct vertex* v2, struct vertex* v3, struct vertex* v4)
{
	int x;
	
	v1->visited = 1;
	v2->visited = 1;
	
	if((v1 == v3 && v2 == v4)||(v1 == v4 && v2 == v3))
		return 1;	
	
	for(x = 0; x < v1->genome_total; x++)
	{
		struct vertex *n1 = v1->vertex_list[x];
		struct vertex *n2 = v2->vertex_list[x];
		
		if(!n1 || !n2)
			continue;	
	
		if(!n1->active || !n2->active)
		{
			printf("shouldn't be checking inactive for cycles\n");
			fflush(stdout);
			exit(0);
		}

		if(x == v1->genome)
			continue;
		
		if(n1->visited || n2->visited)
			continue;
		
		if(n1->genome == n2->genome)
		{
			if(no_cycle_search(n1,n2,v3,v4))
				return 1;
		}
	}
	
	return 0;
}

void clear_visited(struct vertex* v1)
{
	int x;
	
	if(!v1->visited)
		return;
	
	v1->visited = 0;
	
	for(x = 0; x < v1->genome_total; x++)
	{
		struct vertex *t = v1->vertex_list[x];
		
		if(t)
		{
			clear_visited(t);	
		}	
	}
}

int search_perp_v2(struct vertex *v1, struct vertex *v2, struct vertex_controller* vc, int *vertex_count)
{
	int x;
	
	static struct vertex **vertex_memory = NULL;
	
	if(!vertex_memory)
		vertex_memory = (struct vertex**)malloc(sizeof(struct vertex*) * vc->vertex_total * v1->genome_total * 2);
	
	for(x = 0; x < vc->vertex_total * v1->genome_total * 2; x++)
	{
		vertex_memory[x] = NULL;	
	}
	
	static struct vertex ***search_list = NULL;
	if(!search_list)
		search_list = (struct vertex***)malloc(sizeof(struct vertex**) * vc->vertex_total);
	
	for(x = 0; x < vc->vertex_total; x++)
	{
		search_list[x] = &vertex_memory[x*v1->genome_total*2];
	}
	
	int score = sp_v2(v1, v2, 0, search_list, vertex_count);
	
	return score;
}

int sp_v2(struct vertex *v1, struct vertex *v2, int next, struct vertex ***search_list, int *vertex_count)
{
	int x, y, single_gain;
	
	int g = gain(v1,v2);
	int l = loss(v1,v2);
	int r = risk(v1,v2);

	struct vertex *v3, *v4;

	int have = g - l - r;
	int search_slot = 0;
	
	struct graph_modifier gm;
	
	static int max_vertex = 0;
	
	(*vertex_count)++;
	
	//check for cycles in search history
	for(x = 0; x < next; x++)
	{
		for(y = 0; y < v1->genome_total; y++)
		{
			if((search_list[x][y*2] == v1 && search_list[x][y*2+1] == v2)||(search_list[x][y*2] == v2 && search_list[x][y*2+1] == v1))	
			{
				search_list[x][y*2] = NULL;
				search_list[y][y*2+1] = NULL;	
			}
		}
	}
	
	search_list[next][0] = v1;
	search_list[next][1] = v2;
	
	merge_vertex(v1, v2, &gm);
	
	for(x = 0; x < v1->genome_total; x++)
	{		
		v3 = v1->vertex_list[x];
		v4 = v2->vertex_list[x];
		
		if(!v3 || !v4)
			continue;	
	
		if(!v3->active || !v4->active)
			continue;
	
		//printf("step2\n");
	
		if(x == v1->genome)
			continue;
		
		if(v3->genome != v4->genome)
			continue;

		search_list[next][search_slot*2+2] = v3;
		search_list[next][search_slot*2+3] = v4;
	}
	
	for(x = 0; x < v1->genome_total; x++)
	{
		v3 = search_list[next][x*2+2];
		v4 = search_list[next][x*2+3];
		
		if(!v3 || !v4)
			continue;
		
		int from_below = 0;
		single_gain = sp_v2(v3, v4, next+1, search_list, &from_below);
			
		if(single_gain >= 2)
		{
			(*vertex_count)+=from_below;
			have += single_gain;	
		}
	}
	
	replace_vertex(&gm);
	
	for(x = 0; x < v1->genome_total; x++)
	{
		search_list[next][x] = NULL;	
	}

	return have;
}

void search_v3(struct adj_struct** adj_list, int num_genomes)
{
	struct vertex_controller* vc = (struct vertex_controller*)initialize_vertex_controller(10, num_genomes);
	build_graph(adj_list, num_genomes, vc);

	int x, y, z;
	
	int el_count;	
	static struct vertex **event_list = NULL;
	
	if(!event_list)
		event_list = (struct vertex**)malloc(sizeof(struct vertex*) * 2 * vc->vertex_count);	
	for(x = 0; x < vc->vertex_count * 2; x++)
		event_list[x] = NULL;
	
	//printf("mark1 event list sized %d\n", vc->vertex_count * 2);
	//fflush(stdout);
	
	int g;
	
	struct vertex *v1, *v2;
	struct graph_modifier gm;
	
	int flag = 1;
	
	while(flag)
	{
		flag = 0;

		for(x = 0; x < vc->vertex_count - 1; x++)
		{
			v1 = vc->vertex_memory[x];
			
			if(!v1->active)
				continue;
			
			for(y = x+1; y < vc->vertex_count; y++)
			{
				v2 = vc->vertex_memory[y];
				
				if(!v2->active)
					continue;
				
				if(has_perp(v1, v2))
				{	
					int el_count = 0;
					
					g = search_perp_v3(v1, v2, vc, &el_count, event_list);

					if(g > 0)
					{
						//printf("adding splices %d g: %d\n", el_count, g);
						
						for(z = 0; z < el_count; z++)
						{
							struct vertex* m1 = event_list[2*z];
							struct vertex* m2 = event_list[2*z+1];
				
							//printf("splice %d %d\n", m1->terminal, m2->terminal);
				
							merge_vertex(m1, m2, &gm);
							add_adj(adj_list[m1->genome], m1->terminal, m2->terminal);
							flag = 1;

							el_count = 0;
						}
					}
				}
			}
		}
	}
	
	
	for(x = 0; x < num_genomes; x++)
	{
		for(y = 0; y < adj_list[x]->num_genes*2; y++)
		{
			if(adj_list[x]->gene[y] == -1)
			{
				//printf("adding cap\n");
				add_adj(adj_list[x], ITG(y), 0);	
			}	
		}
	}
}

int search_perp_v3(struct vertex *v1, struct vertex *v2, struct vertex_controller *vc, int *event_count, struct vertex **event_list)
{
	int x;
	
	static struct vertex **vertex_memory = NULL;
	
	if(!vertex_memory)
		vertex_memory = (struct vertex**)malloc(sizeof(struct vertex*) * vc->vertex_total * v1->genome_total * 2);
	
	for(x = 0; x < vc->vertex_total * v1->genome_total * 2; x++)
	{
		vertex_memory[x] = NULL;	
	}
	
	static struct vertex ***search_list = NULL;
	if(!search_list)
		search_list = (struct vertex***)malloc(sizeof(struct vertex**) * vc->vertex_total);
	
	for(x = 0; x < vc->vertex_total; x++)
	{
		search_list[x] = &vertex_memory[x*v1->genome_total*2];
	}
	
	int score = spath_v3(v1, v2, 0, search_list, event_count, event_list);
	
	return score;
}

int spath_v3(struct vertex *v1, struct vertex *v2, int next, struct vertex ***search_list, int *event_count, struct vertex **event_list)
{
	int x, y, single_gain;
	
	int g = gain(v1,v2);
	int l = loss(v1,v2);
	int r = risk(v1,v2);

	struct vertex *v3, *v4;

	int have = g - l - r;
	
	//printf("genome: %d t1: %d t2: %d g: %d l: %d r: %d\n", v1->genome, v1->terminal, v2->terminal, g, l, r);
	
	event_list[(*event_count)*2] = v1;
	event_list[(*event_count)*2+1] = v2;
	(*event_count)++;
	
	if(r == 0)
	{
		return have;
	}
	//if(g - l > r)
	//{
		//return have;
	//}
	
	int search_slot = 0;
	
	struct graph_modifier gm;
	
	static int max_vertex = 0;
	
	//printf("recived from parent %d\n", *event_count);
	//fflush(stdout);
	
	//check for cycles in search history
	for(x = 0; x < next; x++)
	{
		for(y = 0; y < v1->genome_total; y++)
		{
			if((search_list[x][y*2] == v1 && search_list[x][y*2+1] == v2)||(search_list[x][y*2] == v2 && search_list[x][y*2+1] == v1))	
			{
				search_list[x][y*2] = NULL;
				search_list[y][y*2+1] = NULL;	
			}
		}
	}
	
	search_list[next][0] = v1;
	search_list[next][1] = v2;
	
	merge_vertex(v1, v2, &gm);
	
	for(x = 0; x < v1->genome_total; x++)
	{		
		v3 = v1->vertex_list[x];
		v4 = v2->vertex_list[x];
		
		if(!v3 || !v4)
			continue;	
	
		if(!v3->active || !v4->active)
			continue;
	
		//printf("step2\n");
	
		if(x == v1->genome)
			continue;
		
		if(v3->genome != v4->genome)
			continue;

		search_list[next][search_slot*2+2] = v3;
		search_list[next][search_slot*2+3] = v4;
	}
	
	for(x = 0; x < v1->genome_total; x++)
	{
		v3 = search_list[next][x*2+2];
		v4 = search_list[next][x*2+3];
		
		if(!v3 || !v4)
			continue;
		
		int from_below = *event_count;
		
		//printf("sending to new branch: %d\n", from_below);
		//fflush(stdout);
		
		single_gain = spath_v3(v3, v4, next+1, search_list, &from_below, event_list);
			
		//printf("recived from finished branch: %d\n", from_below);
		//fflush(stdout);	
			
		if(single_gain >= 2)
		{
			(*event_count)=from_below;
			have += single_gain;	
		}
		else
		{
			//printf("pruning\n");
			//fflush(stdout);	
		}
	}
	
	replace_vertex(&gm);
	
	for(x = 0; x < v1->genome_total; x++)
	{
		search_list[next][x] = NULL;	
	}

	//printf("sending to parent vertex: %d\n", *event_count);
	//fflush(stdout);

	return have;
}