#ifndef VERTEX_H
#define VERTEX_H

#include "structs.h"

struct vertex
{
	int genome;
	int terminal;

	int genome_total;
	
	struct vertex** vertex_list;
	char* odd_list;
	char active;
	
	int* opposite_exit_list;
	
	char visited;
	struct vertex* linked;
};

struct vertex_controller
{
	struct vertex** vertex_memory;
	
	int initial_edges;
	
	int vertex_total;
	int vertex_count;	
};

struct graph_modifier
{
	struct vertex* v1;
	struct vertex* v2;
	
	int double_gain;
	int double_loss;
	int double_potential;
};

struct pair
{
	int gain;
	int vertex_count;
};

struct vertex* alloc_vertex(int genomes);

struct vertex_controller* initialize_vertex_controller(int initial_verticies, int genomes);
struct vertex* get_vertex(struct vertex_controller* vc);
void return_vertex(struct vertex_controller* vc, struct vertex* v);

void clear_visits(struct vertex_controller* vc);

struct vertex* make_vertex(int genome, int terminal, struct vertex_controller* vc);
struct vertex* find_vertex(int genome, int terminal, struct vertex_controller* vc);

//these functions only to be used to construct initial graph
void add_path(struct vertex* v, int exit_genome, char odd);
void add_parallel(struct vertex* v1, int exit1, struct vertex* v2, int exit2);
void add_perpendicular(struct vertex* v1, struct vertex* v2, int exit);

struct vertex* find_vertex(int genome, int terminal, struct vertex_controller* vc);

void build_graph(struct adj_struct** in, int genomes, struct vertex_controller* vc);
void bound_graph(struct vertex_controller* vc, int* upper_bound, int* lower_bound);

void cap_vertex(struct vertex* v, struct graph_modifier* gm);
void merge_vertex(struct vertex* v1, struct vertex* v2, struct graph_modifier* gm);
void replace_vertex(struct graph_modifier* gm);

//-1 = always sub optimal
//0 = potentially optimal
//1 = always optimal
int judge_cap(struct vertex* v);
int judge_splice(struct vertex* v1, struct vertex* v2);
int has_perp(struct vertex* v1, struct vertex* v2);

void print_vertex(struct vertex* v);
void print_graph_modifier(struct graph_modifier* gm);

int find_optimal(struct vertex_controller* vc);
void search_v1(struct vertex_controller* vc, int gain);

int measure_uncertainty(struct vertex** vertex_list, int list_size);
int measure_gain(struct vertex** vertex_list, int list_size);
int measure_gain_recursion(int best, int depth, struct vertex** vertex_list, int list_size);

int gain(struct vertex* v1, struct vertex* v2);
int risk(struct vertex *v1, struct vertex *v2);
int loss(struct vertex *v1, struct vertex *v2);

int search_perp_watch(struct vertex* v1, struct vertex* v2, struct vertex* w1, struct vertex* w2, int depth);

int check_vertex_pair(struct vertex *v1, struct vertex *v2);
int check_opposite_exit(struct vertex *v);

void clear_visited(struct vertex* v1);

int search_perp_v2(struct vertex *v1, struct vertex *v2, struct vertex_controller* vc, int* vertex_count);
int sp_v2(struct vertex *v1, struct vertex *v2, int next, struct vertex ***search_list, int* vertex_count);

int spath_v3(struct vertex *v1, struct vertex *v2, int next, struct vertex ***search_list, int *event_count, struct vertex **event_list);
void search_v3(struct adj_struct** adj_list, int num_genomes);
int search_perp_v3(struct vertex *v1, struct vertex *v2, struct vertex_controller *vc, int *event_count, struct vertex **event_list);


#endif